from __future__ import annotations

import logging
from contextlib import contextmanager
from typing import Iterator, List, Optional, Sequence, Union

from dbt_semantic_interfaces.references import SemanticModelReference
from metricflow_semantics.dag.id_prefix import StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DagId
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.specs.where_filter.where_filter_spec import WhereFilterSpec
from metricflow_semantics.sql.sql_join_type import SqlJoinType

from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.dataflow_plan import (
    DataflowPlan,
    DataflowPlanNode,
)
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor
from metricflow.dataflow.nodes.add_generated_uuid import AddGeneratedUuidColumnNode
from metricflow.dataflow.nodes.aggregate_measures import AggregateMeasuresNode
from metricflow.dataflow.nodes.alias_specs import AliasSpecsNode
from metricflow.dataflow.nodes.combine_aggregated_outputs import CombineAggregatedOutputsNode
from metricflow.dataflow.nodes.compute_metrics import ComputeMetricsNode
from metricflow.dataflow.nodes.constrain_time import ConstrainTimeRangeNode
from metricflow.dataflow.nodes.filter_elements import FilterElementsNode
from metricflow.dataflow.nodes.join_conversion_events import JoinConversionEventsNode
from metricflow.dataflow.nodes.join_over_time import JoinOverTimeRangeNode
from metricflow.dataflow.nodes.join_to_base import JoinOnEntitiesNode
from metricflow.dataflow.nodes.join_to_custom_granularity import JoinToCustomGranularityNode
from metricflow.dataflow.nodes.join_to_time_spine import JoinToTimeSpineNode
from metricflow.dataflow.nodes.metric_time_transform import MetricTimeDimensionTransformNode
from metricflow.dataflow.nodes.min_max import MinMaxNode
from metricflow.dataflow.nodes.order_by_limit import OrderByLimitNode
from metricflow.dataflow.nodes.read_sql_source import ReadSqlSourceNode
from metricflow.dataflow.nodes.semi_additive_join import SemiAdditiveJoinNode
from metricflow.dataflow.nodes.where_filter import WhereConstraintNode
from metricflow.dataflow.nodes.window_reaggregation_node import WindowReaggregationNode
from metricflow.dataflow.nodes.write_to_data_table import WriteToResultDataTableNode
from metricflow.dataflow.nodes.write_to_table import WriteToResultTableNode
from metricflow.dataflow.optimizer.dataflow_plan_optimizer import DataflowPlanOptimizer
from metricflow.dataflow.optimizer.source_scan.source_scan_optimizer import OptimizeBranchResult
from metricflow.plan_conversion.node_processor import PredicateInputType, PredicatePushdownState

logger = logging.getLogger(__name__)


class PredicatePushdownBranchStateTracker:
    """Tracking class for monitoring pushdown state at the node level during a visitor walk."""

    def __init__(self, initial_state: PredicatePushdownState) -> None:  # noqa: D107
        self._current_branch_state: List[PredicatePushdownState] = [initial_state]

    @contextmanager
    def track_pushdown_state(self, pushdown_state: PredicatePushdownState) -> Iterator[None]:
        """Context manager used to track pushdown state along branches in a Dataflow Plan.

        This retains a sequence of pushdown state objects to allow for tracking pushdown opportunities along
        the current branch. Each entry represents the predicate pushdown state as of that point, and as such
        callers need only concern themselves with the value returned by the last_pushdown_state property.

        The back-propagation of pushdown_applied_where_filter_specs is necessary to ensure the outer query
        node can evaluate which where filter specs needs to be applied. We capture the complete set because
        we may have sequenced nodes where some where filters are applied (e.g., time dimension filters might
        be applied to metric time nodes, etc.).

        The state tracking and propagation is built to work as follows:

        For a simple DAG of where_node -> join_node -> source_nodes there will be two branches:

            where_node -> join_node -> left_source_node
            where_node -> join_node -> right_source_node

        In this case the where_node receives the initial predicate pushdown state, and then adds its own
        updated state object (with the where filters) via the context manager and propagates that to the join_node.

        The join_node then receives the where_node's predicate pushdown state, and, for each branch, adds an
        updated state object via the context manager and propagates the updated state objects to the next node.

        The left_source_node gets the join node's left branch state and evaluates it. If it can apply any filters
        it adds an updated state object to note that the filters are applied and propagates it along via the context
        manager. In this case, the context manager exits immediately and returns to the left_source_node, which
        finishes with applying the where constraints and returns back to the join_node.

        At this point, the join node has a left branch context manager with the left_join_branch pushdown state. The
        join_node does NOT have access to the left_source_node's pushdown state, but it needs to be able to notify its
        parent, the where_node, that some filters have been applied at the left_source_node.

        How does it do this? The left_source_node's state update included applied where filters. When the context
        manager exits it propagates the left_source_node's applied where filters back up to the preceding state (in
        this case, the join node's left branch state). The same thing happens when the branch states for the join
        node exit the context manager, so the where_node then sees the union of all filters applied downstream.

        The where_node, then, has access to the complete set of filters applied downstream.

        This is complicated because of joins - we can't store a single set of applied filters, because there's no good
        way to keep them organized in the case of multiple join branches. Instead, we track up and down a single
        branch, and merge the events of parent branches at the join nodes that created them.
        """
        self._current_branch_state.append(pushdown_state)
        yield
        last_visited_pushdown_state = self._current_branch_state.pop(-1)
        if len(last_visited_pushdown_state.applied_where_filter_specs) > 0:
            pushdown_applied_where_filter_specs = frozenset.union(
                *[
                    last_visited_pushdown_state.applied_where_filter_specs,
                    self.last_pushdown_state.applied_where_filter_specs,
                ]
            )
            self.override_last_pushdown_state(
                PredicatePushdownState.with_pushdown_applied_where_filter_specs(
                    original_pushdown_state=self.last_pushdown_state,
                    pushdown_applied_where_filter_specs=pushdown_applied_where_filter_specs,
                )
            )

    @property
    def last_pushdown_state(self) -> PredicatePushdownState:
        """Returns the last seen PredicatePushdownState.

        This is the input state a given node processing method should be using for pushdown operations.
        """
        assert len(self._current_branch_state) > 0, (
            "There should always be at least one element in current branch state! "
            "This suggests an inappropriate removal or improper initialization."
        )
        return self._current_branch_state[-1]

    def override_last_pushdown_state(self, pushdown_state: PredicatePushdownState) -> None:
        """Method for forcibly updating the last seen predicate pushdown state to a new value.

        This is necessary only for cases where we wish to back-propagate some updated state attribute
        for handling in the exit condition of the preceding node in the DAG. The scenario where we use
        this here is for indicating that a where filter has been applied elsewhere on the branch, and so
        outer nodes can skip application as appropriate.

        Since this is not something we want people doing by accident we designate it as a special method
        rather than making it a property setter.
        """
        assert len(self._current_branch_state) > 0, (
            "There should always be at least one element in current branch state! "
            "This suggests an inappropriate removal or improper initialization."
        )
        self._current_branch_state[-1] = pushdown_state


class PredicatePushdownOptimizer(
    DataflowPlanNodeVisitor[OptimizeBranchResult],
    DataflowPlanOptimizer,
):
    """Pushes filter predicates as close to the source node as possible.

    This evaluates filter predicates to determine which, if any, can be directly to an input source node.
    It operates by walking each branch in the DataflowPlan and collecting pushdown state information, then
    evaluating that state at the input source node and applying the filter node (e.g., a WhereConstraintNode)
    directly to the source. As the optimizer unrolls back through the branch it will remove duplicated constraints.

    This assumes that we never do predicate pushdown on a filter that needs to be re-applied, so every filter
    we encounter gets applied exactly once per nested subquery branch encapsulated by a given constraint node.
    """

    def __init__(self, node_data_set_resolver: DataflowPlanNodeOutputDataSetResolver) -> None:
        """Initializer.

        Initializes predicate pushdown state with all optimizer-managed pushdown types enabled, but nothing to
        push down, since time range constraints and where filter specs will be discovered during traversal.
        """
        self._log_level = logging.DEBUG
        self._node_data_set_resolver = node_data_set_resolver
        self._predicate_pushdown_tracker = PredicatePushdownBranchStateTracker(
            initial_state=PredicatePushdownState(
                time_range_constraint=None,
                where_filter_specs=tuple(),
                pushdown_enabled_types=frozenset([PredicateInputType.CATEGORICAL_DIMENSION]),
            )
        )

    def optimize(self, dataflow_plan: DataflowPlan) -> DataflowPlan:  # noqa: D102
        optimized_result: OptimizeBranchResult = dataflow_plan.sink_node.accept(self)

        logger.debug(
            LazyFormat(
                lambda: f"Optimized:\n\n"
                f"{dataflow_plan.sink_node.structure_text()}\n\n"
                f"to:\n\n"
                f"{optimized_result.optimized_branch.structure_text()}",
            ),
        )

        return DataflowPlan(
            plan_id=DagId.from_id_prefix(StaticIdPrefix.OPTIMIZED_DATAFLOW_PLAN_PREFIX),
            sink_nodes=[optimized_result.optimized_branch],
        )

    def _log_visit_node_type(self, node: DataflowPlanNode) -> None:
        logger.debug(
            LazyFormat(
                lambda: f"Visiting {node} with initial pushdown state "
                f"{self._predicate_pushdown_tracker.last_pushdown_state}",
            ),
        )

    def _default_handler(
        self, node: DataflowPlanNode, pushdown_state: Optional[PredicatePushdownState] = None
    ) -> OptimizeBranchResult:
        """Encapsulates state updates and handling for most node types.

        The most common node-level operation is to simply propagate the current predicate pushdown state along and
        return whatever output the parent nodes produce. Of the nodes that do not do this, the most common deviation
        is a pushdown state update.

        As such, this method defaults to propagating the last seen state, but allows an override for cases where
        the handling of the node itself is standard but a pushdown state update is required.
        """
        if pushdown_state is None:
            pushdown_state = self._predicate_pushdown_tracker.last_pushdown_state

        with self._predicate_pushdown_tracker.track_pushdown_state(pushdown_state):
            optimized_parents: Sequence[OptimizeBranchResult] = tuple(
                parent_node.accept(self) for parent_node in node.parent_nodes
            )
            return OptimizeBranchResult(
                optimized_branch=node.with_new_parents(tuple(x.optimized_branch for x in optimized_parents))
            )

    def _models_for_spec(self, spec: WhereFilterSpec) -> Sequence[SemanticModelReference]:
        """Return the distinct semantic models that source the elements referenced in the given where spec.

        TODO: Include special handling for entities, which can be sourced from multiple semantic models
        """
        return tuple(set(element.semantic_model_origin for element in spec.linkable_elements))

    # Source nodes - potential pushdown targets.

    def visit_metric_time_dimension_transform_node(
        self, node: MetricTimeDimensionTransformNode
    ) -> OptimizeBranchResult:
        """Handles predicate pushdown operations against the MetricTimeDimensionTransformNode.

        This node is the one where the metric_time column is constructed. As such, any where filter
        targeting a measure input will be pushed to this node, and no further. In theory we could push
        down directly to the ReadSqlSourceNode, but that requires some juggling on metric_time references
        so we stop here for any matched filters. This shouldn't cause any meaningful problems, as the
        SqlQueryPlanOptimizer processes typically collapse this and the underlying ReadSqlSourceNode into
        a single subquery anyway.

        As this is the base metric_time node, all time-based filter predicate pushdown needs to be managed
        here.

        # TODO: Check relationships between subquery optimizers, time range constraints, and where constraint nodes
        """
        self._log_visit_node_type(node)
        # TODO: Update to handle time range constraints
        return self._push_down_where_filters(node)

    def visit_source_node(self, node: ReadSqlSourceNode) -> OptimizeBranchResult:
        """Handles predicate pushdown to ReadSqlSourceNode.

        This node is currently the root node in the dataflow DAG. In most cases, predicate pushdown will
        stop with the MetricTimeDimensionTransformNode, but if there is ever a scenario where we do a
        metric-free query with a one-sided outer join, an inner join, or a predicate filter set that can
        be pushed past outer join boundaries (via some kind of semantic analysis or other semantic guarantee)
        we want to make sure any applicable filters can be bound as closely to the input source as possible.
        """
        self._log_visit_node_type(node)
        return self._push_down_where_filters(node)

    def _push_down_where_filters(
        self, node: Union[MetricTimeDimensionTransformNode, ReadSqlSourceNode]
    ) -> OptimizeBranchResult:
        """Helper method for pushing where filters down to base source nodes.

        This only accepts the two supported source node types - the ReadSqlSourceNode and the
        MetricTimeDimensionTransformNode. In theory we could push down to ReadSqlSourceNode in every scenario, but
        in practice this gets tricky given that filters on metric_time are expected, and metric_time is not a column
        available on the original sql source node.
        """
        current_pushdown_state = self._predicate_pushdown_tracker.last_pushdown_state
        node_semantic_models = node.as_plan().source_semantic_models
        if len(node_semantic_models) != 1 or not current_pushdown_state.has_where_filters_to_push_down:
            return self._default_handler(node)

        source_semantic_model, *_ = node_semantic_models
        filters_to_apply: List[WhereFilterSpec] = []
        filters_left_over: List[WhereFilterSpec] = []
        source_node_linkable_specs = self._node_data_set_resolver.get_output_data_set(
            node
        ).instance_set.spec_set.linkable_specs

        for filter_spec in current_pushdown_state.where_filter_specs:
            filter_spec_semantic_models = self._models_for_spec(filter_spec)
            invalid_element_types = [
                element
                for element in filter_spec.linkable_elements
                if element.element_type not in current_pushdown_state.pushdown_eligible_element_types
            ]
            if len(filter_spec_semantic_models) != 1 or len(invalid_element_types) > 0:
                continue

            all_linkable_specs_match = all(spec in source_node_linkable_specs for spec in filter_spec.linkable_specs)
            # TODO: Handle the case where entities can be defined in multiple models, only one of which need match
            semantic_models_match = filter_spec_semantic_models[0] == source_semantic_model
            if all_linkable_specs_match and semantic_models_match:
                filters_to_apply.append(filter_spec)
            else:
                filters_left_over.append(filter_spec)

        logger.debug(LazyFormat(lambda: f"Filter specs to add:\n{filters_to_apply}"))

        applied_filters = frozenset.union(
            *[frozenset(current_pushdown_state.applied_where_filter_specs), frozenset(filters_to_apply)]
        )
        updated_pushdown_state = PredicatePushdownState(
            time_range_constraint=current_pushdown_state.time_range_constraint,
            where_filter_specs=tuple(filters_left_over),
            pushdown_enabled_types=current_pushdown_state.pushdown_enabled_types,
            applied_where_filter_specs=applied_filters,
        )
        optimized_node = self._default_handler(node=node, pushdown_state=updated_pushdown_state)
        if len(filters_to_apply) > 0:
            return OptimizeBranchResult(
                optimized_branch=WhereConstraintNode.create(
                    parent_node=optimized_node.optimized_branch, where_specs=filters_to_apply
                )
            )
        else:
            return optimized_node

    # Constraint nodes - predicate sources for pushdown.

    def visit_constrain_time_range_node(self, node: ConstrainTimeRangeNode) -> OptimizeBranchResult:
        """Adds time range constraint information from the input node to the current pushdown state.

        For now we simply overwrite the window with the time range constraint value we find here. In future
        we may wish to allow for a set of time constraints that we can union over at filter time.
        """
        self._log_visit_node_type(node)
        updated_pushdown_state = PredicatePushdownState.with_time_range_constraint(
            original_pushdown_state=self._predicate_pushdown_tracker.last_pushdown_state,
            time_range_constraint=node.time_range_constraint,
        )
        return self._default_handler(node=node, pushdown_state=updated_pushdown_state)

    def visit_where_constraint_node(self, node: WhereConstraintNode) -> OptimizeBranchResult:
        """Adds where filters from the input node to the current pushdown state.

        The WhereConstraintNode carries the filter information in the form of WhereFilterSpecs, which may or may
        not be eligible for pushdown. This processor simply propagates them forward so long as where filter
        predicate pushdown is still enabled for this branch.

        When the visitor returns to this node from its parents, it updates the pushdown state for this node in the
        tracker. It does this within the scope of the context manager in order to keep the pushdown state updates
        consistent - modifying only the state entry associated with this node, and allowing the tracker to do all
        of the upstream state propagation.

        The fact that the filters have been added at this point does not mean they will be pushed down, as intervening
        join nodes might remove them from consideration, so we must ensure all filters are applied as specified
        within this method.
        """
        self._log_visit_node_type(node)
        current_pushdown_state = self._predicate_pushdown_tracker.last_pushdown_state
        if not current_pushdown_state.where_filter_pushdown_enabled:
            return self._default_handler(node)

        updated_pushdown_state = PredicatePushdownState.with_where_filter_specs(
            original_pushdown_state=current_pushdown_state,
            where_filter_specs=tuple(current_pushdown_state.where_filter_specs) + tuple(node.input_where_specs),
        )

        with self._predicate_pushdown_tracker.track_pushdown_state(updated_pushdown_state):
            optimized_parent: OptimizeBranchResult = node.parent_node.accept(self)
            pushdown_state_updated_by_parent = self._predicate_pushdown_tracker.last_pushdown_state
            applied_filter_specs = pushdown_state_updated_by_parent.applied_where_filter_specs
            filter_specs_to_apply = [spec for spec in node.input_where_specs if spec not in applied_filter_specs]
            if len(applied_filter_specs) > 0:
                updated_specs = frozenset.union(
                    frozenset(node.input_where_specs),
                    pushdown_state_updated_by_parent.applied_where_filter_specs,
                )
                self._predicate_pushdown_tracker.override_last_pushdown_state(
                    PredicatePushdownState.with_pushdown_applied_where_filter_specs(
                        original_pushdown_state=pushdown_state_updated_by_parent,
                        pushdown_applied_where_filter_specs=updated_specs,
                    )
                )
                logger.debug(
                    LazyFormat(
                        lambda: f"Added applied specs to pushdown state. Added specs:\n\n{node.input_where_specs}\n\n"
                        + f"Updated pushdown state:\n\n{self._predicate_pushdown_tracker.last_pushdown_state}"
                    ),
                )

            if node.always_apply:
                logger.debug(
                    LazyFormat(
                        lambda: f"Applying original filter spec set based on node-level override directive. "
                        f"Additional specs "
                        f"appled:\n{[spec for spec in node.input_where_specs if spec not in filter_specs_to_apply]}"
                    ),
                )
                optimized_node = OptimizeBranchResult(
                    optimized_branch=node.with_new_parents((optimized_parent.optimized_branch,))
                )
            elif len(filter_specs_to_apply) > 0:
                optimized_node = OptimizeBranchResult(
                    optimized_branch=WhereConstraintNode.create(
                        parent_node=optimized_parent.optimized_branch, where_specs=filter_specs_to_apply
                    )
                )
            else:
                optimized_node = optimized_parent

        return optimized_node

    # Join nodes - these may affect pushdown state based on join type

    def visit_combine_aggregated_outputs_node(self, node: CombineAggregatedOutputsNode) -> OptimizeBranchResult:
        """Removes where filter specs from current pushdown state while allowing subsequent specs to be pushed down.

        The combine aggregated outputs node typically does a FULL OUTER JOIN, which means any where constraint applied
        after it cannot be safely pushed down if it might include NULLs. However, where constraints applied
        to parents of this node can still be pushed down along the branch, and time range constraints will never allow
        NULLs to pass so those should remain intact.

        There is a CROSS JOIN case which would enable pushdown, but accessing that scenario requires rebuilding
        the dataset for this node, and as this is rarely, if ever, wrapped by a WhereConstraintNode we don't bother
        handling this scenario at this time.
        """
        self._log_visit_node_type(node)
        updated_pushdown_state = PredicatePushdownState.without_where_filter_specs(
            original_pushdown_state=self._predicate_pushdown_tracker.last_pushdown_state,
        )
        return self._default_handler(node=node, pushdown_state=updated_pushdown_state)

    def visit_join_conversion_events_node(self, node: JoinConversionEventsNode) -> OptimizeBranchResult:
        """Updates predicate pushdown state in a manner appropriate for the JoinConversionEventsNode.

        As of right now the JoinConversionEvents node does some wonky stuff with filter expressions. More broadly, it
        is not entirely clear if it's ok to push an arbitrary filter expression down past this point in the graph,
        because an outside filter might be meant to apply to the conversion output rather than the conversion input,
        and managing the conversion window and time filters gets tricky.

        TODO: Enable predicate pushdown once we establish clear expectations for conversion metric filter behaviors.

        """
        self._log_visit_node_type(node)
        base_node_pushdown_state = PredicatePushdownState.without_where_filter_specs(
            original_pushdown_state=self._predicate_pushdown_tracker.last_pushdown_state,
        )
        # The conversion metric branch silently removes all filters, so this is a redundant operation.
        # TODO: Enable pushdown for the conversion metric branch when filters are supported
        conversion_node_pushdown_state = PredicatePushdownState.with_pushdown_disabled()

        optimized_parents: List[OptimizeBranchResult] = []
        with self._predicate_pushdown_tracker.track_pushdown_state(base_node_pushdown_state):
            optimized_parents.append(node.base_node.accept(self))

        with self._predicate_pushdown_tracker.track_pushdown_state(conversion_node_pushdown_state):
            optimized_parents.append(node.conversion_node.accept(self))

        return OptimizeBranchResult(
            optimized_branch=node.with_new_parents(
                new_parent_nodes=tuple(x.optimized_branch for x in optimized_parents)
            )
        )

    def visit_join_to_custom_granularity_node(  # noqa: D102
        self, node: JoinToCustomGranularityNode
    ) -> OptimizeBranchResult:
        raise NotImplementedError

    def visit_alias_specs_node(self, node: AliasSpecsNode) -> OptimizeBranchResult:  # noqa: D102
        raise NotImplementedError

    def visit_join_on_entities_node(self, node: JoinOnEntitiesNode) -> OptimizeBranchResult:
        """Handles pushdown state propagation for the standard join node type.

        This node type has two sets of parent nodes - a left node and a set of join targets - and the pushdown state
        must be updated separately for each parent based on the relevant join type. What's more, each parent represents
        a branch in the DAG, and as such the state propagation must happen independently for each.

        In particular, if a given branch is ever a target of any OUTER JOIN we cannot safely push down any filter that
        might allow for a NULL value, as that affects query semantics. In other words, the left branch cannot allow
        predicate pushdown for these filter types if there is even a single FULL OUTER or RIGHT OUTER JOIN in the
        target list.

        Note - at this time we only apply time constraints to measure nodes, and those are always on the left side
        of the join. As such, time constraints are not propagated to the right side of the join. This restriction
        may be relaxed at a later time, but for now it is largely irrelevant since we do not allow fanout joins and
        do not yet have support for pre-filters based on time ranges for things like SCD joins.

        Note we initialize branch state tracking objects prior to traversal to avoid back-propagation from
        one branch affecting the predicate pushdown behavior along other branches.
        """
        self._log_visit_node_type(node)
        left_parent = node.left_node
        if any(join_description.join_type is SqlJoinType.FULL_OUTER for join_description in node.join_targets):
            left_branch_pushdown_state = PredicatePushdownState.without_where_filter_specs(
                original_pushdown_state=self._predicate_pushdown_tracker.last_pushdown_state,
            )
        else:
            left_branch_pushdown_state = self._predicate_pushdown_tracker.last_pushdown_state

        base_right_branch_pushdown_state = PredicatePushdownState.without_time_range_constraint(
            self._predicate_pushdown_tracker.last_pushdown_state
        )
        outer_join_right_branch_pushdown_state = PredicatePushdownState.without_where_filter_specs(
            original_pushdown_state=base_right_branch_pushdown_state
        )

        optimized_parents: List[OptimizeBranchResult] = []

        with self._predicate_pushdown_tracker.track_pushdown_state(left_branch_pushdown_state):
            optimized_parents.append(left_parent.accept(self))

        for join_description in node.join_targets:
            if (
                join_description.join_type is SqlJoinType.LEFT_OUTER
                or join_description.join_type is SqlJoinType.FULL_OUTER
            ):
                right_branch_pushdown_state = outer_join_right_branch_pushdown_state
            else:
                right_branch_pushdown_state = base_right_branch_pushdown_state
            with self._predicate_pushdown_tracker.track_pushdown_state(right_branch_pushdown_state):
                optimized_parents.append(join_description.join_node.accept(self))

        return OptimizeBranchResult(
            optimized_branch=node.with_new_parents(tuple(x.optimized_branch for x in optimized_parents))
        )

    def visit_join_over_time_range_node(self, node: JoinOverTimeRangeNode) -> OptimizeBranchResult:
        """Updates time range constraint window to account for join over time range behavior, as needed.

        For the time being we simply pass through in all cases, because time constraint adjustments are
        handled in the DataflowPlanBuilder and the original time constraint is passed through. This handler
        will need refinement once we move support for time range constraints and time filters to this class.

        Note the join type is always INNER, which means we only need special handling for filters involving
        time spans.

        TODO: move constraint window adjustment to the optimizer for application of the relevant TimeRangeConstraint.
        """
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_join_to_time_spine_node(self, node: JoinToTimeSpineNode) -> OptimizeBranchResult:
        """Updates pushdown state to account for time spine join types and constraint window requirements.

        The JoinToTimeSpineNode is processed as the left node in a join against a single input parent, typically
        a measure or metric input of some kind, but the join type depends on how the JoinToTimeSpineNode is configured.
        Since we cannot safely push down filters that might allow for nulls, we must update the pushdown state
        accordingly.

        TODO: move constraint window management to the optimizer and enable time-based filter propagation
        as appropriate
        """
        self._log_visit_node_type(node)
        current_pushdown_state = self._predicate_pushdown_tracker.last_pushdown_state
        if node.join_type is SqlJoinType.LEFT_OUTER or node.join_type is SqlJoinType.FULL_OUTER:
            updated_pushdown_state = PredicatePushdownState(
                time_range_constraint=None,
                where_filter_specs=tuple(),
                pushdown_enabled_types=current_pushdown_state.pushdown_enabled_types,
                applied_where_filter_specs=current_pushdown_state.applied_where_filter_specs,
            )
        else:
            updated_pushdown_state = PredicatePushdownState.without_time_range_constraint(current_pushdown_state)

        return self._default_handler(node=node, pushdown_state=updated_pushdown_state)

    # Nodes affecting pushdown state for other reasons

    def visit_compute_metrics_node(self, node: ComputeMetricsNode) -> OptimizeBranchResult:
        """Disables or refines pushdown state for metrics that require custom handling.

        For example, time filters for derived offset or cumulative metrics might require special state
        handling for filters on time dimension inputs or time range constraint expressions.
        """
        self._log_visit_node_type(node)
        current_pushdown_state = self._predicate_pushdown_tracker.last_pushdown_state
        if any(metric_spec.has_time_offset for metric_spec in node.metric_specs):
            # TODO: Allow non-time filters for offset metrics. This is for parity with the original hook preventing
            # invalid pushdown for offset metrics
            updated_pushdown_state = PredicatePushdownState.with_pushdown_disabled()
        else:
            updated_pushdown_state = current_pushdown_state

        return self._default_handler(node=node, pushdown_state=updated_pushdown_state)

    # Other nodes - these simply propagate state, as they do not affect predicate pushdown in our context

    def visit_aggregate_measures_node(self, node: AggregateMeasuresNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_filter_elements_node(self, node: FilterElementsNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_order_by_limit_node(self, node: OrderByLimitNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_semi_additive_join_node(self, node: SemiAdditiveJoinNode) -> OptimizeBranchResult:
        """Propagates pushdown state to all input branches.

        The semi-additive join node only does inner joins, so it does not affect pushdown state.
        """
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_write_to_result_data_table_node(  # noqa: D102
        self, node: WriteToResultDataTableNode
    ) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_write_to_result_table_node(self, node: WriteToResultTableNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_add_generated_uuid_column_node(  # noqa: D102
        self, node: AddGeneratedUuidColumnNode
    ) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_min_max_node(self, node: MinMaxNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_handler(node)

    def visit_window_reaggregation_node(self, node: WindowReaggregationNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_handler(node)



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import List, Tuple

from metricflow.sql.sql_plan import SqlSelectColumn

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SelectColumnSet:
    """A set of SQL select columns that represent the different instance types in a data set."""

    metric_columns: List[SqlSelectColumn] = field(default_factory=list)
    measure_columns: List[SqlSelectColumn] = field(default_factory=list)
    dimension_columns: List[SqlSelectColumn] = field(default_factory=list)
    time_dimension_columns: List[SqlSelectColumn] = field(default_factory=list)
    entity_columns: List[SqlSelectColumn] = field(default_factory=list)
    group_by_metric_columns: List[SqlSelectColumn] = field(default_factory=list)
    metadata_columns: List[SqlSelectColumn] = field(default_factory=list)

    def merge(self, other_set: SelectColumnSet) -> SelectColumnSet:
        """Combine the select columns by type."""
        return SelectColumnSet(
            metric_columns=self.metric_columns + other_set.metric_columns,
            measure_columns=self.measure_columns + other_set.measure_columns,
            dimension_columns=self.dimension_columns + other_set.dimension_columns,
            time_dimension_columns=self.time_dimension_columns + other_set.time_dimension_columns,
            entity_columns=self.entity_columns + other_set.entity_columns,
            group_by_metric_columns=self.group_by_metric_columns + other_set.group_by_metric_columns,
            metadata_columns=self.metadata_columns + other_set.metadata_columns,
        )

    def as_tuple(self) -> Tuple[SqlSelectColumn, ...]:
        """Return all select columns as a tuple."""
        return tuple(
            # This order was chosen to match the column sequence data consumers typically prefer.
            self.time_dimension_columns
            + self.entity_columns
            + self.dimension_columns
            + self.group_by_metric_columns
            + self.metric_columns
            + self.measure_columns
            + self.metadata_columns
        )

    def without_measure_columns(self) -> SelectColumnSet:
        """Returns this but with the measure columns removed."""
        return SelectColumnSet(
            metric_columns=self.metric_columns,
            dimension_columns=self.dimension_columns,
            time_dimension_columns=self.time_dimension_columns,
            entity_columns=self.entity_columns,
            group_by_metric_columns=self.group_by_metric_columns,
            metadata_columns=self.metadata_columns,
        )



from __future__ import annotations

import dataclasses
import itertools
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, List, Sequence, Tuple, TypeVar

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from typing_extensions import override

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.specs.instance_spec import InstanceSpec, InstanceSpecVisitor, LinkableInstanceSpec

if TYPE_CHECKING:
    from metricflow_semantics.specs.dimension_spec import DimensionSpec
    from metricflow_semantics.specs.entity_spec import EntitySpec
    from metricflow_semantics.specs.group_by_metric_spec import GroupByMetricSpec
    from metricflow_semantics.specs.measure_spec import MeasureSpec
    from metricflow_semantics.specs.metadata_spec import MetadataSpec
    from metricflow_semantics.specs.metric_spec import MetricSpec
    from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec


@dataclass(frozen=True)
class InstanceSpecSet(Mergeable, SerializableDataclass):
    """Consolidates all specs used in an instance set."""

    metric_specs: Tuple[MetricSpec, ...] = ()
    measure_specs: Tuple[MeasureSpec, ...] = ()
    dimension_specs: Tuple[DimensionSpec, ...] = ()
    entity_specs: Tuple[EntitySpec, ...] = ()
    time_dimension_specs: Tuple[TimeDimensionSpec, ...] = ()
    group_by_metric_specs: Tuple[GroupByMetricSpec, ...] = ()
    metadata_specs: Tuple[MetadataSpec, ...] = ()

    @override
    def merge(self, other: InstanceSpecSet) -> InstanceSpecSet:
        return InstanceSpecSet(
            metric_specs=self.metric_specs + other.metric_specs,
            measure_specs=self.measure_specs + other.measure_specs,
            dimension_specs=self.dimension_specs + other.dimension_specs,
            entity_specs=self.entity_specs + other.entity_specs,
            group_by_metric_specs=self.group_by_metric_specs + other.group_by_metric_specs,
            time_dimension_specs=self.time_dimension_specs + other.time_dimension_specs,
            metadata_specs=self.metadata_specs + other.metadata_specs,
        )

    @override
    @classmethod
    def empty_instance(cls) -> InstanceSpecSet:
        return InstanceSpecSet()

    def dedupe(self) -> InstanceSpecSet:
        """De-duplicates repeated elements.

        TBD: Have merge de-duplicate instead.
        """
        metric_specs_deduped = []
        for metric_spec in self.metric_specs:
            if metric_spec not in metric_specs_deduped:
                metric_specs_deduped.append(metric_spec)

        measure_specs_deduped = []
        for measure_spec in self.measure_specs:
            if measure_spec not in measure_specs_deduped:
                measure_specs_deduped.append(measure_spec)

        dimension_specs_deduped = []
        for dimension_spec in self.dimension_specs:
            if dimension_spec not in dimension_specs_deduped:
                dimension_specs_deduped.append(dimension_spec)

        time_dimension_specs_deduped = []
        for time_dimension_spec in self.time_dimension_specs:
            if time_dimension_spec not in time_dimension_specs_deduped:
                time_dimension_specs_deduped.append(time_dimension_spec)

        entity_specs_deduped = []
        for entity_spec in self.entity_specs:
            if entity_spec not in entity_specs_deduped:
                entity_specs_deduped.append(entity_spec)

        group_by_metric_specs_deduped = []
        for group_by_metric_spec in self.group_by_metric_specs:
            if group_by_metric_spec not in group_by_metric_specs_deduped:
                group_by_metric_specs_deduped.append(group_by_metric_spec)

        return InstanceSpecSet(
            metric_specs=tuple(metric_specs_deduped),
            measure_specs=tuple(measure_specs_deduped),
            dimension_specs=tuple(dimension_specs_deduped),
            time_dimension_specs=tuple(time_dimension_specs_deduped),
            entity_specs=tuple(entity_specs_deduped),
            group_by_metric_specs=tuple(group_by_metric_specs_deduped),
        )

    @property
    def linkable_specs(self) -> Sequence[LinkableInstanceSpec]:
        """All linkable specs in this set."""
        return list(
            itertools.chain(
                self.dimension_specs, self.time_dimension_specs, self.entity_specs, self.group_by_metric_specs
            )
        )

    @property
    def all_specs(self) -> Sequence[InstanceSpec]:  # noqa: D102
        return tuple(
            itertools.chain(
                self.measure_specs,
                self.dimension_specs,
                self.time_dimension_specs,
                self.entity_specs,
                self.group_by_metric_specs,
                self.metric_specs,
                self.metadata_specs,
            )
        )

    def transform(  # noqa: D102
        self, transform_function: InstanceSpecSetTransform[TransformOutputT]
    ) -> TransformOutputT:
        return transform_function.transform(self)

    @property
    def metric_time_specs(self) -> Sequence[TimeDimensionSpec]:
        """Returns any specs referring to metric time at any grain."""
        return tuple(
            time_dimension_spec
            for time_dimension_spec in self.time_dimension_specs
            if time_dimension_spec.is_metric_time
        )

    @staticmethod
    def create_from_specs(specs: Sequence[InstanceSpec]) -> InstanceSpecSet:  # noqa: D102
        return group_specs_by_type(specs)


TransformOutputT = TypeVar("TransformOutputT")


class InstanceSpecSetTransform(Generic[TransformOutputT], ABC):
    """Function to use for transforming spec sets."""

    @abstractmethod
    def transform(self, spec_set: InstanceSpecSet) -> TransformOutputT:  # noqa: D102
        pass


@dataclass
class _GroupSpecByTypeVisitor(InstanceSpecVisitor[None]):
    """Groups a spec by type into an `InstanceSpecSet`."""

    metric_specs: List[MetricSpec] = dataclasses.field(default_factory=list)
    measure_specs: List[MeasureSpec] = dataclasses.field(default_factory=list)
    dimension_specs: List[DimensionSpec] = dataclasses.field(default_factory=list)
    entity_specs: List[EntitySpec] = dataclasses.field(default_factory=list)
    time_dimension_specs: List[TimeDimensionSpec] = dataclasses.field(default_factory=list)
    group_by_metric_specs: List[GroupByMetricSpec] = dataclasses.field(default_factory=list)
    metadata_specs: List[MetadataSpec] = dataclasses.field(default_factory=list)

    @override
    def visit_measure_spec(self, measure_spec: MeasureSpec) -> None:
        self.measure_specs.append(measure_spec)

    @override
    def visit_dimension_spec(self, dimension_spec: DimensionSpec) -> None:
        self.dimension_specs.append(dimension_spec)

    @override
    def visit_time_dimension_spec(self, time_dimension_spec: TimeDimensionSpec) -> None:
        self.time_dimension_specs.append(time_dimension_spec)

    @override
    def visit_entity_spec(self, entity_spec: EntitySpec) -> None:
        self.entity_specs.append(entity_spec)

    @override
    def visit_group_by_metric_spec(self, group_by_metric_spec: GroupByMetricSpec) -> None:
        self.group_by_metric_specs.append(group_by_metric_spec)

    @override
    def visit_metric_spec(self, metric_spec: MetricSpec) -> None:
        self.metric_specs.append(metric_spec)

    @override
    def visit_metadata_spec(self, metadata_spec: MetadataSpec) -> None:
        self.metadata_specs.append(metadata_spec)


def group_specs_by_type(specs: Sequence[InstanceSpec]) -> InstanceSpecSet:
    """Groups a sequence of specs by type."""
    grouper = _GroupSpecByTypeVisitor()
    for spec in specs:
        spec.accept(grouper)

    return InstanceSpecSet(
        metric_specs=tuple(grouper.metric_specs),
        measure_specs=tuple(grouper.measure_specs),
        dimension_specs=tuple(grouper.dimension_specs),
        entity_specs=tuple(grouper.entity_specs),
        time_dimension_specs=tuple(grouper.time_dimension_specs),
        group_by_metric_specs=tuple(grouper.group_by_metric_specs),
        metadata_specs=tuple(grouper.metadata_specs),
    )


def group_spec_by_type(spec: InstanceSpec) -> InstanceSpecSet:
    """Similar to group_specs_by_type() but for a single spec."""
    return group_specs_by_type((spec,))



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Sequence, Tuple

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.specs.entity_spec import LinklessEntitySpec
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.sql.sql_join_type import SqlJoinType
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.builder.partitions import (
    PartitionDimensionJoinDescription,
    PartitionTimeDimensionJoinDescription,
)
from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True)
class ValidityWindowJoinDescription:
    """Encapsulates details about join constraints around validity windows."""

    window_start_dimension: TimeDimensionSpec
    window_end_dimension: TimeDimensionSpec


@dataclass(frozen=True)
class JoinDescription:
    """Describes how data from a node should be joined to data from another node."""

    join_node: DataflowPlanNode
    join_on_entity: Optional[LinklessEntitySpec]
    join_type: SqlJoinType

    join_on_partition_dimensions: Tuple[PartitionDimensionJoinDescription, ...]
    join_on_partition_time_dimensions: Tuple[PartitionTimeDimensionJoinDescription, ...]

    validity_window: Optional[ValidityWindowJoinDescription] = None

    def __post_init__(self) -> None:  # noqa: D105
        if self.join_on_entity is None and self.join_type != SqlJoinType.CROSS_JOIN:
            raise RuntimeError("`join_on_entity` is required unless using CROSS JOIN.")


@dataclass(frozen=True, eq=False)
class JoinOnEntitiesNode(DataflowPlanNode):
    """A node that joins data from other nodes via the entities in the inputs.

    Attributes:
        left_node: Node with standard output.
        join_targets: Other sources that should be joined to this node.
    """

    left_node: DataflowPlanNode
    join_targets: Tuple[JoinDescription, ...]

    @staticmethod
    def create(  # noqa: D102
        left_node: DataflowPlanNode,
        join_targets: Sequence[JoinDescription],
    ) -> JoinOnEntitiesNode:
        parent_nodes = [left_node] + [join_target.join_node for join_target in join_targets]
        return JoinOnEntitiesNode(
            parent_nodes=tuple(parent_nodes),
            left_node=left_node,
            join_targets=tuple(join_targets),
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_JOIN_TO_STANDARD_OUTPUT_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_join_on_entities_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return """Join Standard Outputs"""

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + tuple(
            DisplayedProperty(f"join{i}_for_node_id_{join_description.join_node.node_id}", join_description)
            for i, join_description in enumerate(self.join_targets)
        )

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        if not isinstance(other_node, self.__class__) or len(self.join_targets) != len(other_node.join_targets):
            return False

        for i in range(len(self.join_targets)):
            if (
                self.join_targets[i].join_on_entity != other_node.join_targets[i].join_on_entity
                or self.join_targets[i].join_on_partition_dimensions
                != other_node.join_targets[i].join_on_partition_dimensions
                or self.join_targets[i].join_on_partition_time_dimensions
                != other_node.join_targets[i].join_on_partition_time_dimensions
                or self.join_targets[i].validity_window != other_node.join_targets[i].validity_window
            ):
                return False
        return True

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> JoinOnEntitiesNode:  # noqa: D102
        assert len(new_parent_nodes) > 1
        new_left_node = new_parent_nodes[0]
        new_join_nodes = new_parent_nodes[1:]
        assert len(new_join_nodes) == len(self.join_targets)

        return JoinOnEntitiesNode.create(
            left_node=new_left_node,
            join_targets=[
                JoinDescription(
                    join_node=new_join_nodes[i],
                    join_on_entity=old_join_target.join_on_entity,
                    join_on_partition_dimensions=old_join_target.join_on_partition_dimensions,
                    join_on_partition_time_dimensions=old_join_target.join_on_partition_time_dimensions,
                    validity_window=old_join_target.validity_window,
                    join_type=old_join_target.join_type,
                )
                for i, old_join_target in enumerate(self.join_targets)
            ],
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass

from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec


@dataclass(frozen=True)
class PartitionSpecSet(SerializableDataclass):
    """Grouping of the linkable specs."""

    dimension_specs: Tuple[DimensionSpec, ...] = ()
    time_dimension_specs: Tuple[TimeDimensionSpec, ...] = ()



from __future__ import annotations

import typing
from typing import List, Sequence

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.protocols import Metric, WhereFilterIntersection
from dbt_semantic_interfaces.references import (
    MeasureReference,
    MetricReference,
)
from dbt_semantic_interfaces.type_enums import MetricType
from typing_extensions import override

from metricflow_semantics.model.linkable_element_property import LinkableElementProperty
from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.model.semantics.element_filter import LinkableElementFilter
from metricflow_semantics.model.semantics.linkable_element_set import LinkableElementSet
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryResolutionIssue,
    MetricFlowQueryResolutionIssueSet,
)
from metricflow_semantics.query.issues.parsing.cumulative_metric_requires_metric_time import (
    CumulativeMetricRequiresMetricTimeIssue,
)
from metricflow_semantics.query.issues.parsing.offset_metric_requires_metric_time import (
    OffsetMetricRequiresMetricTimeIssue,
)
from metricflow_semantics.query.issues.parsing.scd_requires_metric_time import (
    ScdRequiresMetricTimeIssue,
)
from metricflow_semantics.query.resolver_inputs.query_resolver_inputs import ResolverInputForQuery
from metricflow_semantics.query.validation_rules.base_validation_rule import PostResolutionQueryValidationRule

if typing.TYPE_CHECKING:
    from metricflow_semantics.query.query_resolver import ResolveGroupByItemsResult


class MetricTimeQueryValidationRule(PostResolutionQueryValidationRule):
    """Validates cases where a query requires metric_time to be specified as a group-by-item.

    Currently, known cases are:

    * Cumulative metrics.
    * Derived metrics with an offset time.
    * Slowly changing dimensions
    """

    def __init__(  # noqa: D107
        self,
        manifest_lookup: SemanticManifestLookup,
        resolver_input_for_query: ResolverInputForQuery,
        resolve_group_by_item_result: ResolveGroupByItemsResult,
    ) -> None:
        super().__init__(
            manifest_lookup=manifest_lookup,
            resolver_input_for_query=resolver_input_for_query,
            resolve_group_by_item_result=resolve_group_by_item_result,
        )

        self._query_includes_metric_time = (
            self._resolve_group_by_item_result.linkable_element_set.filter(
                LinkableElementFilter(with_any_of=frozenset({LinkableElementProperty.METRIC_TIME}))
            ).spec_count
            > 0
        )

        self._scd_linkable_element_set = self._resolve_group_by_item_result.linkable_element_set.filter(
            LinkableElementFilter(with_any_of=frozenset({LinkableElementProperty.SCD_HOP}))
        )

    def _query_includes_agg_time_dimension_of_metric(self, metric_reference: MetricReference) -> bool:
        valid_agg_time_dimensions = self._manifest_lookup.metric_lookup.get_valid_agg_time_dimensions_for_metric(
            metric_reference
        )
        return (
            len(set(valid_agg_time_dimensions).intersection(self._resolve_group_by_item_result.group_by_item_specs)) > 0
        )

    def _validate_cumulative_metric(
        self,
        metric_reference: MetricReference,
        metric: Metric,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> Sequence[MetricFlowQueryResolutionIssue]:
        # A cumulative metric with a window or grain-to-date specified requires a `metric-time` or the aggregation time
        # dimension for the metric.
        if (
            metric.type_params is not None
            and metric.type_params.cumulative_type_params is not None
            and (
                metric.type_params.cumulative_type_params.window is not None
                or metric.type_params.cumulative_type_params.grain_to_date is not None
            )
        ):
            if self._query_includes_metric_time or self._query_includes_agg_time_dimension_of_metric(metric_reference):
                return ()

            return (
                CumulativeMetricRequiresMetricTimeIssue.from_parameters(
                    metric_reference=metric_reference,
                    query_resolution_path=resolution_path,
                ),
            )
        return ()

    def _validate_derived_metric(
        self,
        metric_reference: MetricReference,
        metric: Metric,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> Sequence[MetricFlowQueryResolutionIssue]:
        has_time_offset = any(
            input_metric.offset_window is not None or input_metric.offset_to_grain is not None
            for input_metric in metric.input_metrics
        )

        # If a derived metric does not define a time offset, then there are no requirements on what's in the group-by
        # items.
        if not has_time_offset:
            return ()

        # If a derived metric has a time offset, then the query needs to include `metric_time` or the aggregation time
        # dimension of a metric.
        if self._query_includes_metric_time or self._query_includes_agg_time_dimension_of_metric(metric_reference):
            return ()

        return (
            OffsetMetricRequiresMetricTimeIssue.from_parameters(
                metric_reference=metric_reference,
                input_metrics=metric.input_metrics,
                query_resolution_path=resolution_path,
            ),
        )

    def _scd_linkable_element_set_for_measure(self, measure_reference: MeasureReference) -> LinkableElementSet:
        """Returns subset of the query's `LinkableElements` that are SCDs and associated with the measure."""
        measure_semantic_model = self._manifest_lookup.semantic_model_lookup.measure_lookup.get_properties(
            measure_reference
        ).model_reference

        return self._scd_linkable_element_set.filter_by_left_semantic_model(measure_semantic_model)

    @override
    def validate_metric_in_resolution_dag(
        self,
        metric_reference: MetricReference,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> MetricFlowQueryResolutionIssueSet:
        metric = self._manifest_lookup.metric_lookup.get_metric(metric_reference)
        issues: List[MetricFlowQueryResolutionIssue] = []

        if metric.type is MetricType.CUMULATIVE:
            issues.extend(
                self._validate_cumulative_metric(
                    metric_reference=metric_reference,
                    metric=metric,
                    resolution_path=resolution_path,
                )
            )

        elif metric.type is MetricType.RATIO or metric.type is MetricType.DERIVED:
            issues.extend(
                self._validate_derived_metric(
                    metric_reference=metric_reference,
                    metric=metric,
                    resolution_path=resolution_path,
                )
            )
        elif metric.type is MetricType.SIMPLE:
            pass
        elif metric.type is MetricType.CONVERSION:
            pass
        else:
            assert_values_exhausted(metric.type)

        return MetricFlowQueryResolutionIssueSet(issues=tuple(issues))

    @override
    def validate_query_in_resolution_dag(
        self,
        metrics_in_query: Sequence[MetricReference],
        where_filter_intersection: WhereFilterIntersection,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> MetricFlowQueryResolutionIssueSet:
        return MetricFlowQueryResolutionIssueSet.empty_instance()

    @override
    def validate_measure_in_resolution_dag(
        self,
        measure_reference: MeasureReference,
        resolution_path: MetricFlowQueryResolutionPath,
    ) -> MetricFlowQueryResolutionIssueSet:
        scd_linkable_elemenent_set_for_measure = self._scd_linkable_element_set_for_measure(measure_reference)

        if scd_linkable_elemenent_set_for_measure.spec_count == 0:
            return MetricFlowQueryResolutionIssueSet.empty_instance()

        if self._query_includes_metric_time:
            return MetricFlowQueryResolutionIssueSet.empty_instance()

        # Queries that join to an SCD don't support direct references to agg_time_dimension, so we
        # only check for metric_time. If we decide to support agg_time_dimension, we should add a check

        return MetricFlowQueryResolutionIssueSet.from_issue(
            ScdRequiresMetricTimeIssue.from_parameters(
                scds_in_query=scd_linkable_elemenent_set_for_measure.specs, query_resolution_path=resolution_path
            )
        )



from __future__ import annotations

import dataclasses
import itertools
import typing
from dataclasses import dataclass
from typing import Dict, List, Sequence, Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.references import MeasureReference, MetricReference
from typing_extensions import override

from metricflow_semantics.collection_helpers.merger import Mergeable
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.entity_spec import EntitySpec
from metricflow_semantics.specs.group_by_metric_spec import GroupByMetricSpec
from metricflow_semantics.specs.instance_spec import InstanceSpecVisitor, LinkableInstanceSpec
from metricflow_semantics.specs.spec_set import InstanceSpecSet
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec

if typing.TYPE_CHECKING:
    from metricflow_semantics.model.semantics.metric_lookup import MetricLookup
    from metricflow_semantics.model.semantics.semantic_model_lookup import SemanticModelLookup
    from metricflow_semantics.specs.measure_spec import MeasureSpec
    from metricflow_semantics.specs.metadata_spec import MetadataSpec
    from metricflow_semantics.specs.metric_spec import MetricSpec


@dataclass(frozen=True)
class LinkableSpecSet(Mergeable, SerializableDataclass):
    """Groups linkable specs."""

    dimension_specs: Tuple[DimensionSpec, ...] = ()
    time_dimension_specs: Tuple[TimeDimensionSpec, ...] = ()
    entity_specs: Tuple[EntitySpec, ...] = ()
    group_by_metric_specs: Tuple[GroupByMetricSpec, ...] = ()

    @property
    def contains_metric_time(self) -> bool:
        """Returns true if this set contains a spec referring to metric time at any grain."""
        return len(self.metric_time_specs) > 0

    @property
    def time_dimension_specs_with_custom_grain(self) -> Tuple[TimeDimensionSpec, ...]:  # noqa: D102
        return tuple([spec for spec in self.time_dimension_specs if spec.time_granularity.is_custom_granularity])

    def replace_custom_granularity_with_base_granularity(self) -> LinkableSpecSet:
        """Return the same spec set, replacing any custom time granularity with its base granularity."""
        return LinkableSpecSet(
            dimension_specs=self.dimension_specs,
            time_dimension_specs=tuple(
                [time_dimension_spec.with_base_grain() for time_dimension_spec in self.time_dimension_specs]
            ),
            entity_specs=self.entity_specs,
            group_by_metric_specs=self.group_by_metric_specs,
        )

    def included_agg_time_dimension_specs_for_metric(
        self, metric_reference: MetricReference, metric_lookup: MetricLookup
    ) -> List[TimeDimensionSpec]:
        """Get the time dims included that are valid agg time dimensions for the specified metric."""
        queried_metric_time_specs = list(self.metric_time_specs)

        valid_agg_time_dimensions = metric_lookup.get_valid_agg_time_dimensions_for_metric(metric_reference)
        queried_agg_time_dimension_specs = (
            list(set(self.time_dimension_specs).intersection(set(valid_agg_time_dimensions)))
            + queried_metric_time_specs
        )

        return queried_agg_time_dimension_specs

    def included_agg_time_dimension_specs_for_measure(
        self, measure_reference: MeasureReference, semantic_model_lookup: SemanticModelLookup
    ) -> List[TimeDimensionSpec]:
        """Get the time dims included that are valid agg time dimensions for the specified measure."""
        queried_metric_time_specs = list(self.metric_time_specs)

        valid_agg_time_dimensions = semantic_model_lookup.measure_lookup.get_properties(
            measure_reference
        ).agg_time_dimension_specs
        queried_agg_time_dimension_specs = (
            list(set(self.time_dimension_specs).intersection(set(valid_agg_time_dimensions)))
            + queried_metric_time_specs
        )

        return queried_agg_time_dimension_specs

    @property
    def metric_time_specs(self) -> Sequence[TimeDimensionSpec]:
        """Returns any specs referring to metric time at any grain."""
        return tuple(
            time_dimension_spec
            for time_dimension_spec in self.time_dimension_specs
            if time_dimension_spec.is_metric_time
        )

    @property
    def as_tuple(self) -> Tuple[LinkableInstanceSpec, ...]:  # noqa: D102
        return tuple(
            itertools.chain(
                self.dimension_specs, self.time_dimension_specs, self.entity_specs, self.group_by_metric_specs
            )
        )

    def add_specs(
        self,
        dimension_specs: Tuple[DimensionSpec, ...] = (),
        time_dimension_specs: Tuple[TimeDimensionSpec, ...] = (),
        entity_specs: Tuple[EntitySpec, ...] = (),
        group_by_metric_specs: Tuple[GroupByMetricSpec, ...] = (),
    ) -> LinkableSpecSet:
        """Return a new set with the new specs in addition to the existing ones."""
        return LinkableSpecSet(
            dimension_specs=self.dimension_specs + dimension_specs,
            time_dimension_specs=self.time_dimension_specs + time_dimension_specs,
            entity_specs=self.entity_specs + entity_specs,
            group_by_metric_specs=self.group_by_metric_specs + group_by_metric_specs,
        )

    @override
    def merge(self, other: LinkableSpecSet) -> LinkableSpecSet:
        return self.add_specs(
            dimension_specs=other.dimension_specs,
            time_dimension_specs=other.time_dimension_specs,
            entity_specs=other.entity_specs,
            group_by_metric_specs=other.group_by_metric_specs,
        )

    @classmethod
    @override
    def empty_instance(cls) -> LinkableSpecSet:
        return LinkableSpecSet()

    def dedupe(self) -> LinkableSpecSet:  # noqa: D102
        # Use dictionaries to dedupe as it preserves insertion order.

        dimension_spec_dict: Dict[DimensionSpec, None] = {}
        for dimension_spec in self.dimension_specs:
            dimension_spec_dict[dimension_spec] = None

        time_dimension_spec_dict: Dict[TimeDimensionSpec, None] = {}
        for time_dimension_spec in self.time_dimension_specs:
            time_dimension_spec_dict[time_dimension_spec] = None

        entity_spec_dict: Dict[EntitySpec, None] = {}
        for entity_spec in self.entity_specs:
            entity_spec_dict[entity_spec] = None

        group_by_metric_spec_dict: Dict[GroupByMetricSpec, None] = {}
        for group_by_metric in self.group_by_metric_specs:
            group_by_metric_spec_dict[group_by_metric] = None

        return LinkableSpecSet(
            dimension_specs=tuple(dimension_spec_dict.keys()),
            time_dimension_specs=tuple(time_dimension_spec_dict.keys()),
            entity_specs=tuple(entity_spec_dict.keys()),
            group_by_metric_specs=tuple(group_by_metric_spec_dict.keys()),
        )

    def is_subset_of(self, other_set: LinkableSpecSet) -> bool:  # noqa: D102
        return set(self.as_tuple).issubset(set(other_set.as_tuple))

    def difference(self, other: LinkableSpecSet) -> LinkableSpecSet:  # noqa: D102
        return LinkableSpecSet(
            dimension_specs=tuple(set(self.dimension_specs) - set(other.dimension_specs)),
            time_dimension_specs=tuple(set(self.time_dimension_specs) - set(other.time_dimension_specs)),
            entity_specs=tuple(set(self.entity_specs) - set(other.entity_specs)),
            group_by_metric_specs=tuple(set(self.group_by_metric_specs) - set(other.group_by_metric_specs)),
        )

    @staticmethod
    def create_from_specs(specs: Sequence[LinkableInstanceSpec]) -> LinkableSpecSet:  # noqa: D102
        return _group_specs_by_type(specs)

    @property
    def as_instance_spec_set(self) -> InstanceSpecSet:  # noqa: D102
        return InstanceSpecSet(
            dimension_specs=self.dimension_specs,
            entity_specs=self.entity_specs,
            time_dimension_specs=self.time_dimension_specs,
            group_by_metric_specs=self.group_by_metric_specs,
        )


@dataclass
class _GroupSpecByTypeVisitor(InstanceSpecVisitor[None]):
    """Groups a spec by type into an `InstanceSpecSet`."""

    dimension_specs: List[DimensionSpec] = dataclasses.field(default_factory=list)
    entity_specs: List[EntitySpec] = dataclasses.field(default_factory=list)
    time_dimension_specs: List[TimeDimensionSpec] = dataclasses.field(default_factory=list)
    group_by_metric_specs: List[GroupByMetricSpec] = dataclasses.field(default_factory=list)

    @override
    def visit_measure_spec(self, measure_spec: MeasureSpec) -> None:
        pass

    @override
    def visit_dimension_spec(self, dimension_spec: DimensionSpec) -> None:
        self.dimension_specs.append(dimension_spec)

    @override
    def visit_time_dimension_spec(self, time_dimension_spec: TimeDimensionSpec) -> None:
        self.time_dimension_specs.append(time_dimension_spec)

    @override
    def visit_entity_spec(self, entity_spec: EntitySpec) -> None:
        self.entity_specs.append(entity_spec)

    @override
    def visit_group_by_metric_spec(self, group_by_metric_spec: GroupByMetricSpec) -> None:
        self.group_by_metric_specs.append(group_by_metric_spec)

    @override
    def visit_metric_spec(self, metric_spec: MetricSpec) -> None:
        pass

    @override
    def visit_metadata_spec(self, metadata_spec: MetadataSpec) -> None:
        pass


def _group_specs_by_type(specs: Sequence[LinkableInstanceSpec]) -> LinkableSpecSet:
    """Groups a sequence of specs by type."""
    grouper = _GroupSpecByTypeVisitor()
    for spec in specs:
        spec.accept(grouper)

    return LinkableSpecSet(
        dimension_specs=tuple(grouper.dimension_specs),
        entity_specs=tuple(grouper.entity_specs),
        time_dimension_specs=tuple(grouper.time_dimension_specs),
        group_by_metric_specs=tuple(grouper.group_by_metric_specs),
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Sequence, Tuple

from typing_extensions import override

from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.naming.object_builder_scheme import ObjectBuilderNamingScheme
from metricflow_semantics.query.group_by_item.candidate_push_down.group_by_item_candidate import GroupByItemCandidateSet
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.base_node import GroupByItemResolutionNode
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput


@dataclass(frozen=True)
class NoCommonItemsInParents(MetricFlowQueryResolutionIssue):
    """Describes an issue during group-by-item resolution where parents don't have common items that match a pattern.

    Currently, this only occurs when resolving a time dimension specified without a grain in a where filter. In such
    cases, only the time dimension at the defined grain is available from parents. This is how we defined how such
    ambiguous time dimensions should be resolved.

    e.g. a query for a daily metric and a monthly metric would consist of a query node with the daily metric node and
    the monthly metric node as the parents.

    To resolve a query filter like "{{ TimeDimension('metric_time') }} = '2020-01-01'", the query node would receive
    metric_time__day from the daily metric node and metric_time__month from the monthly metric node. Then this issue
    would be created to let the user know that the ambiguous group-by-item in the where filter can't be resolved.
    """

    parent_candidate_sets: Tuple[GroupByItemCandidateSet, ...]

    @staticmethod
    def from_parameters(  # noqa: D102
        query_resolution_path: MetricFlowQueryResolutionPath,
        parent_node_to_candidate_set: Dict[GroupByItemResolutionNode, GroupByItemCandidateSet],
        parent_issues: Sequence[MetricFlowQueryResolutionIssue],
    ) -> NoCommonItemsInParents:
        return NoCommonItemsInParents(
            issue_type=MetricFlowQueryIssueType.ERROR,
            query_resolution_path=query_resolution_path,
            parent_candidate_sets=tuple(candidate_set for _, candidate_set in parent_node_to_candidate_set.items()),
            parent_issues=tuple(parent_issues),
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        last_path_item = self.query_resolution_path.last_item
        last_path_item_parent_descriptions = ", ".join(
            [parent_node.ui_description for parent_node in last_path_item.parent_nodes]
        )
        naming_scheme = (
            associated_input.input_pattern_description.naming_scheme
            if associated_input.input_pattern_description is not None
            else ObjectBuilderNamingScheme()
        )

        parent_to_available_items = {}
        for candidate_set in self.parent_candidate_sets:
            resolution_node = candidate_set.path_from_leaf_node.last_item
            spec_as_strs = tuple(naming_scheme.input_str(spec) for spec in candidate_set.specs)
            parent_to_available_items["Matching items for: " + resolution_node.ui_description] = [
                (spec_str if spec_str is not None else "None") for spec_str in spec_as_strs
            ]
        return (
            f"{last_path_item.ui_description} is built from:\n\n"
            f"{indent(last_path_item_parent_descriptions)}.\n"
            f"However, the given input does not match to a common item that is available to those parents:\n\n"
            f"{indent(mf_pformat(parent_to_available_items, max_line_length=80))}\n\n"
            f"For time dimension inputs, please specify a time grain as ambiguous resolution only allows "
            f"resolution when the parents have the same defined time gain."
        )

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> NoCommonItemsInParents:
        return NoCommonItemsInParents(
            issue_type=self.issue_type,
            parent_issues=tuple(issue.with_path_prefix(path_prefix) for issue in self.parent_issues),
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
            parent_candidate_sets=tuple(
                candidate_set.with_path_prefix(path_prefix) for candidate_set in self.parent_candidate_sets
            ),
        )



from __future__ import annotations

import logging
from functools import cached_property
from typing import Dict, List, Optional, Sequence, Set, Tuple

from dbt_semantic_interfaces.protocols.entity import Entity
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.protocols.semantic_model import SemanticModel
from dbt_semantic_interfaces.references import (
    DimensionReference,
    EntityReference,
    MeasureReference,
    SemanticModelElementReference,
    SemanticModelReference,
    TimeDimensionReference,
)
from dbt_semantic_interfaces.type_enums import DimensionType

from metricflow_semantics.errors.error_classes import InvalidSemanticModelError
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.model.semantics.dimension_lookup import DimensionLookup
from metricflow_semantics.model.semantics.element_group import ElementGrouper
from metricflow_semantics.model.semantics.measure_lookup import MeasureLookup
from metricflow_semantics.model.spec_converters import MeasureConverter
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.entity_spec import EntitySpec
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec
from metricflow_semantics.specs.measure_spec import MeasureSpec
from metricflow_semantics.specs.non_additive_dimension_spec import NonAdditiveDimensionSpec
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec
from metricflow_semantics.time.granularity import ExpandedTimeGranularity

logger = logging.getLogger(__name__)


class SemanticModelLookup:
    """Tracks semantic information for semantic models held in a set of SemanticModelContainers."""

    def __init__(self, model: SemanticManifest, custom_granularities: Dict[str, ExpandedTimeGranularity]) -> None:
        """Initializer.

        Args:
            model: the semantic manifest used for loading semantic model definitions
        """
        self._custom_granularities = custom_granularities
        self._measure_index: Dict[MeasureReference, SemanticModel] = {}
        self._measure_non_additive_dimension_specs: Dict[MeasureReference, NonAdditiveDimensionSpec] = {}
        self._dimension_index: Dict[DimensionReference, List[SemanticModel]] = {}
        self._entity_index: Dict[EntityReference, List[SemanticModel]] = {}

        self._dimension_ref_to_spec: Dict[DimensionReference, DimensionSpec] = {}
        self._entity_ref_to_spec: Dict[EntityReference, EntitySpec] = {}

        self._semantic_model_to_aggregation_time_dimensions: Dict[
            SemanticModelReference, ElementGrouper[TimeDimensionReference, MeasureSpec]
        ] = {}

        self._semantic_model_reference_to_semantic_model: Dict[SemanticModelReference, SemanticModel] = {}
        sorted_semantic_models = sorted(model.semantic_models, key=lambda semantic_model: semantic_model.name)
        for semantic_model in sorted_semantic_models:
            self._add_semantic_model(semantic_model)

        self._measure_lookup = MeasureLookup(sorted_semantic_models, custom_granularities)
        self._dimension_lookup = DimensionLookup(sorted_semantic_models)

    @cached_property
    def custom_granularity_names(self) -> Tuple[str, ...]:
        """Returns all the custom_granularity names."""
        return tuple(self._custom_granularities.keys())

    def get_dimension_references(self) -> Sequence[DimensionReference]:
        """Retrieve all dimension references from the collection of semantic models."""
        return tuple(self._dimension_index.keys())

    @property
    def measure_references(self) -> Sequence[MeasureReference]:
        """Return all measure references from the collection of semantic models."""
        return list(self._measure_index.keys())

    @property
    def non_additive_dimension_specs_by_measure(self) -> Dict[MeasureReference, NonAdditiveDimensionSpec]:
        """Return a mapping from all semi-additive measures to their corresponding non additive dimension parameters.

        This includes all measures with non-additive dimension parameters, if any, from the collection of semantic models.
        """
        return self._measure_non_additive_dimension_specs

    def get_entity_references(self) -> Sequence[EntityReference]:
        """Retrieve all entity references from the collection of semantic models."""
        return list(self._entity_index.keys())

    def get_entity_in_semantic_model(self, ref: SemanticModelElementReference) -> Optional[Entity]:
        """Retrieve the entity matching the element -> semantic model mapping, if any."""
        semantic_model = self.get_by_reference(ref.semantic_model_reference)
        if not semantic_model:
            return None

        for entity in semantic_model.entities:
            if entity.reference.element_name == ref.element_name:
                return entity

        return None

    def get_by_reference(self, semantic_model_reference: SemanticModelReference) -> Optional[SemanticModel]:
        """Retrieve the semantic model object matching the input semantic model reference, if any."""
        return self._semantic_model_reference_to_semantic_model.get(semantic_model_reference)

    def _add_semantic_model(self, semantic_model: SemanticModel) -> None:
        """Add semantic model semantic information, validating consistency with existing semantic models."""
        errors = []

        if semantic_model.reference in self._semantic_model_reference_to_semantic_model:
            errors.append(f"Semantic model {semantic_model.reference} already added.")

        if len(errors) > 0:
            raise InvalidSemanticModelError(f"Error adding {semantic_model.reference}. Got errors: {errors}")

        self._semantic_model_to_aggregation_time_dimensions[semantic_model.reference] = ElementGrouper[
            TimeDimensionReference, MeasureSpec
        ]()

        for measure in semantic_model.measures:
            self._measure_index[measure.reference] = semantic_model
            agg_time_dimension_reference = semantic_model.checked_agg_time_dimension_for_measure(measure.reference)

            matching_dimensions = tuple(
                dimension
                for dimension in semantic_model.dimensions
                if dimension.type == DimensionType.TIME
                and dimension.time_dimension_reference == agg_time_dimension_reference
            )

            matching_dimensions_count = len(matching_dimensions)
            if matching_dimensions_count != 1:
                raise RuntimeError(
                    f"Found {matching_dimensions_count} matching dimensions for {agg_time_dimension_reference} "
                    f"in {semantic_model}"
                )
            agg_time_dimension = matching_dimensions[0]
            if agg_time_dimension.type_params is None or agg_time_dimension.type_params.time_granularity is None:
                raise RuntimeError(
                    f"Aggregation time dimension does not have a time granularity set: {agg_time_dimension}"
                )

            self._semantic_model_to_aggregation_time_dimensions[semantic_model.reference].add_value(
                key=TimeDimensionReference(
                    element_name=agg_time_dimension.name,
                ),
                value=MeasureConverter.convert_to_measure_spec(measure=measure),
            )

            if measure.non_additive_dimension:
                non_additive_dimension_spec = NonAdditiveDimensionSpec(
                    name=measure.non_additive_dimension.name,
                    window_choice=measure.non_additive_dimension.window_choice,
                    window_groupings=tuple(measure.non_additive_dimension.window_groupings),
                )
                self._measure_non_additive_dimension_specs[measure.reference] = non_additive_dimension_spec
        for dim in semantic_model.dimensions:
            semantic_models_for_dimension = self._dimension_index.get(dim.reference, []) + [semantic_model]
            self._dimension_index[dim.reference] = semantic_models_for_dimension

            if not StructuredLinkableSpecName.from_name(
                qualified_name=dim.name, custom_granularity_names=self.custom_granularity_names
            ).is_element_name:
                # TODO: [custom granularity] change this to an assertion once we're sure there aren't exceptions
                logger.warning(
                    LazyFormat(
                        lambda: f"Dimension name `{dim.name}` contains annotations, but this name should be the plain element name "
                        "from the original model. This should have been blocked by validation!"
                    )
                )

            # TODO: Construct these specs correctly. All of the time dimension specs have the default granularity
            self._dimension_ref_to_spec[dim.time_dimension_reference or dim.reference] = (
                TimeDimensionSpec(element_name=dim.name, entity_links=())
                if dim.type is DimensionType.TIME
                else DimensionSpec(element_name=dim.name, entity_links=())
            )

        for entity in semantic_model.entities:
            semantic_models_for_entity = self._entity_index.get(entity.reference, []) + [semantic_model]
            self._entity_index[entity.reference] = semantic_models_for_entity

            self._entity_ref_to_spec[entity.reference] = EntitySpec(element_name=entity.name, entity_links=())

        self._semantic_model_reference_to_semantic_model[semantic_model.reference] = semantic_model

    def get_aggregation_time_dimensions_with_measures(
        self, semantic_model_reference: SemanticModelReference
    ) -> ElementGrouper[TimeDimensionReference, MeasureSpec]:
        """Return all time dimensions in a semantic model with their associated measures."""
        assert (
            semantic_model_reference in self._semantic_model_to_aggregation_time_dimensions
        ), f"Semantic Model {semantic_model_reference} is not known"
        return self._semantic_model_to_aggregation_time_dimensions[semantic_model_reference]

    def get_semantic_models_for_entity(self, entity_reference: EntityReference) -> Set[SemanticModel]:
        """Return all semantic models associated with an entity reference."""
        return set(self._entity_index.get(entity_reference, []))

    def get_semantic_models_for_dimension(self, dimension_reference: DimensionReference) -> Set[SemanticModel]:
        """Return all semantic models associated with a dimension reference."""
        return set(self._dimension_index.get(dimension_reference, []))

    def get_element_spec_for_name(self, element_name: str) -> LinkableInstanceSpec:
        """Returns the spec for the given name of a linkable element (dimension or entity)."""
        if TimeDimensionReference(element_name=element_name) in self._dimension_ref_to_spec:
            return self._dimension_ref_to_spec[TimeDimensionReference(element_name=element_name)]
        elif DimensionReference(element_name=element_name) in self._dimension_ref_to_spec:
            return self._dimension_ref_to_spec[DimensionReference(element_name=element_name)]
        elif EntityReference(element_name=element_name) in self._entity_ref_to_spec:
            return self._entity_ref_to_spec[EntityReference(element_name=element_name)]
        else:
            raise ValueError(f"Unable to find linkable element {element_name} in manifest")

    @property
    def measure_lookup(self) -> MeasureLookup:  # noqa: D102
        return self._measure_lookup

    @property
    def dimension_lookup(self) -> DimensionLookup:  # noqa: D102
        return self._dimension_lookup



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence, Tuple

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.specs.measure_spec import MetricInputMeasureSpec
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class AggregateMeasuresNode(DataflowPlanNode):
    """A node that aggregates the measures by the associated group by elements.

    In the event that one or more of the aggregated input measures has an alias assigned to it, any output query
    resulting from an operation on this node must apply the alias and transform the measure instances accordingly,
    otherwise this join could produce a query with two identically named measure columns with, e.g., different
    constraints applied to the measure.

    The input measure specs are required for downstream nodes to be aware of any input measures with
    user-provided aliases, such as we might encounter with constrained and unconstrained versions of the
    same input measure.
    """

    metric_input_measure_specs: Tuple[MetricInputMeasureSpec, ...]

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(  # noqa: D102
        parent_node: DataflowPlanNode, metric_input_measure_specs: Sequence[MetricInputMeasureSpec]
    ) -> AggregateMeasuresNode:
        return AggregateMeasuresNode(
            parent_nodes=(parent_node,), metric_input_measure_specs=tuple(metric_input_measure_specs)
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_AGGREGATE_MEASURES_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_aggregate_measures_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return """Aggregate Measures"""

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return (
            isinstance(other_node, self.__class__)
            and other_node.metric_input_measure_specs == self.metric_input_measure_specs
        )

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> AggregateMeasuresNode:  # noqa: D102
        assert len(new_parent_nodes) == 1
        return AggregateMeasuresNode(
            parent_nodes=tuple(new_parent_nodes),
            metric_input_measure_specs=self.metric_input_measure_specs,
        )



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence

from metricflow_semantics.dag.id_prefix import StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DagId
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat

from metricflow.dataflow.dataflow_plan import (
    DataflowPlan,
    DataflowPlanNode,
)
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor
from metricflow.dataflow.nodes.add_generated_uuid import AddGeneratedUuidColumnNode
from metricflow.dataflow.nodes.aggregate_measures import AggregateMeasuresNode
from metricflow.dataflow.nodes.alias_specs import AliasSpecsNode
from metricflow.dataflow.nodes.combine_aggregated_outputs import CombineAggregatedOutputsNode
from metricflow.dataflow.nodes.compute_metrics import ComputeMetricsNode
from metricflow.dataflow.nodes.constrain_time import ConstrainTimeRangeNode
from metricflow.dataflow.nodes.filter_elements import FilterElementsNode
from metricflow.dataflow.nodes.join_conversion_events import JoinConversionEventsNode
from metricflow.dataflow.nodes.join_over_time import JoinOverTimeRangeNode
from metricflow.dataflow.nodes.join_to_base import JoinOnEntitiesNode
from metricflow.dataflow.nodes.join_to_custom_granularity import JoinToCustomGranularityNode
from metricflow.dataflow.nodes.join_to_time_spine import JoinToTimeSpineNode
from metricflow.dataflow.nodes.metric_time_transform import MetricTimeDimensionTransformNode
from metricflow.dataflow.nodes.min_max import MinMaxNode
from metricflow.dataflow.nodes.order_by_limit import OrderByLimitNode
from metricflow.dataflow.nodes.read_sql_source import ReadSqlSourceNode
from metricflow.dataflow.nodes.semi_additive_join import SemiAdditiveJoinNode
from metricflow.dataflow.nodes.where_filter import WhereConstraintNode
from metricflow.dataflow.nodes.window_reaggregation_node import WindowReaggregationNode
from metricflow.dataflow.nodes.write_to_data_table import WriteToResultDataTableNode
from metricflow.dataflow.nodes.write_to_table import WriteToResultTableNode
from metricflow.dataflow.optimizer.dataflow_plan_optimizer import DataflowPlanOptimizer
from metricflow.dataflow.optimizer.source_scan.cm_branch_combiner import (
    ComputeMetricsBranchCombiner,
    ComputeMetricsBranchCombinerResult,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class OptimizeBranchResult:  # noqa: D101
    optimized_branch: DataflowPlanNode


@dataclass(frozen=True)
class BranchCombinationResult:
    """Holds the results of combining a branch (right_branch) with one of the branches in a list (left_branch)."""

    left_branch: DataflowPlanNode
    right_branch: DataflowPlanNode
    combined_branch: Optional[DataflowPlanNode] = None


class SourceScanOptimizer(
    DataflowPlanNodeVisitor[OptimizeBranchResult],
    DataflowPlanOptimizer,
):
    """Reduces the number of scans (ReadSqlSourceNodes) in a dataflow plan.

    This attempts to reduce the number of scans by combining the parent nodes of CombineAggregatedOutputsNode via the
    ComputeMetricsBranchCombiner.

    A plan with a structure similar to
        ...
        <CombineAggregatedOutputsNode>
            <ComputeMetricsNode metrics="[metric0]">
                <AggregateMeasuresNode>
                ...
                </AggregateMeasuresNode>
            </ComputeMetricsNode>
            <ComputeMetricsNode metrics="[metrics1]">
                <AggregateMeasuresNode>
                ...
                </AggregateMeasuresNode>
            </ComputeMetricsNode>
            <ComputeMetricsNode metric="[metrics2]">
                <AggregateMeasuresNode>
                ...
                </AggregateMeasuresNode>
            </ComputeMetricsNode>
        </CombineAggregatedOutputsNode>
        ...
    will be converted to
        ...
        <CombineAggregatedOutputsNode>
            <ComputeMetricsNode metrics="[metric0, metric1]">
                <AggregateMeasuresNode>
                ...
                </AggregateMeasuresNode>
            </ComputeMetricsNode>
            <ComputeMetricsNode metrics="[metric2]">
                <AggregateMeasuresNode>
                ...
                </AggregateMeasuresNode>
            </ComputeMetricsNode>
        </CombineAggregatedOutputsNode>
        ...
    when possible.

    In cases where all ComputeMetricsNodes can be combined into a single one, the CombineAggregatedOutputsNode may be removed as
    well.

    This traverses the dataflow plan using DFS. When visiting a node (current_node), it first runs the optimization
    process on the parent branches, then tries to create a more optimal version of the current_node with those optimized
    parents.
    """

    def __init__(self) -> None:  # noqa: D107
        self._node_to_result: Dict[DataflowPlanNode, OptimizeBranchResult] = {}

    def _log_visit_node_type(self, node: DataflowPlanNode) -> None:
        logger.debug(LazyFormat(lambda: f"Visiting {node.node_id}"))

    def _default_base_output_handler(
        self,
        node: DataflowPlanNode,
    ) -> OptimizeBranchResult:
        memoized_result = self._node_to_result.get(node)
        if memoized_result is not None:
            return memoized_result

        optimized_parent_nodes: Sequence[DataflowPlanNode] = tuple(
            parent_node.accept(self).optimized_branch for parent_node in node.parent_nodes
        )

        # If no optimization is done, use the same nodes so that common operations can be identified for CTE generation.
        if tuple(node.parent_nodes) == optimized_parent_nodes:
            result = OptimizeBranchResult(optimized_branch=node)
        else:
            result = OptimizeBranchResult(optimized_branch=node.with_new_parents(optimized_parent_nodes))

        self._node_to_result[node] = result
        return result

    def visit_source_node(self, node: ReadSqlSourceNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_join_on_entities_node(self, node: JoinOnEntitiesNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_aggregate_measures_node(self, node: AggregateMeasuresNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_window_reaggregation_node(self, node: WindowReaggregationNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_compute_metrics_node(self, node: ComputeMetricsNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        # Run the optimizer on the parent branch to handle derived metrics, which are defined recursively in the DAG.

        memoized_result = self._node_to_result.get(node)
        if memoized_result is not None:
            return memoized_result

        optimized_parent_result: OptimizeBranchResult = node.parent_node.accept(self)
        if optimized_parent_result.optimized_branch is not None:
            result = OptimizeBranchResult(
                optimized_branch=ComputeMetricsNode.create(
                    parent_node=optimized_parent_result.optimized_branch,
                    metric_specs=node.metric_specs,
                    for_group_by_source_node=node.for_group_by_source_node,
                    aggregated_to_elements=node.aggregated_to_elements,
                )
            )
        else:
            result = OptimizeBranchResult(optimized_branch=node)

        self._node_to_result[node] = result
        return result

    def visit_order_by_limit_node(self, node: OrderByLimitNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_where_constraint_node(self, node: WhereConstraintNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_write_to_result_data_table_node(  # noqa: D102
        self, node: WriteToResultDataTableNode
    ) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_write_to_result_table_node(self, node: WriteToResultTableNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_filter_elements_node(self, node: FilterElementsNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    @staticmethod
    def _combine_branches(
        left_branches: Sequence[DataflowPlanNode], right_branch: DataflowPlanNode
    ) -> Sequence[BranchCombinationResult]:
        """Combine the right branch with one of the left branches.

        This is intended to be used in a loop where the goal is to combine a set of branches with each other in the most
        optimal way. This should be the case if the combination of branches is commutative e.g. if combining branches
        (a + b) + c is the same as a + (b + c).
        """
        results = []
        combined = False
        for left_branch in left_branches:
            # Try combining only if we haven't combined before.
            if not combined:
                combiner = ComputeMetricsBranchCombiner(left_branch_node=left_branch)
                combiner_result: ComputeMetricsBranchCombinerResult = right_branch.accept(combiner)
                if combiner_result.combined_branch is not None:
                    combined = True
                    results.append(
                        BranchCombinationResult(
                            left_branch=left_branch,
                            right_branch=right_branch,
                            combined_branch=combiner_result.combined_branch,
                        )
                    )
                    continue

            # Add to results if we didn't combine with left_branch
            results.append(
                BranchCombinationResult(
                    left_branch=left_branch,
                    right_branch=right_branch,
                )
            )
        return results

    def visit_combine_aggregated_outputs_node(  # noqa: D102
        self, node: CombineAggregatedOutputsNode
    ) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)

        memoized_result = self._node_to_result.get(node)
        if memoized_result is not None:
            return memoized_result

        # The parent node of the CombineAggregatedOutputsNode can be either ComputeMetricsNodes or
        # CombineAggregatedOutputsNodes.
        # Stores the result of running this optimizer on each parent branch separately.
        optimized_parent_branches = []
        logger.debug(LazyFormat(lambda: f"{node.node_id} has {len(node.parent_nodes)} parent branches"))

        # Run the optimizer on the parent branch to handle derived metrics, which are defined recursively in the DAG.
        for parent_branch in node.parent_nodes:
            result: OptimizeBranchResult = parent_branch.accept(self)
            optimized_parent_branches.append(result.optimized_branch)

        # Try to combine (using ComputeMetricsBranchCombiner) as many parent branches as possible in a
        # greedy N^2 approach. The optimality of this approach needs more thought to prove conclusively, but given
        # the seemingly transitive properties of the combination operation, this seems reasonable.
        combined_parent_branches: List[DataflowPlanNode] = []
        for optimized_parent_branch in optimized_parent_branches:
            combination_results = SourceScanOptimizer._combine_branches(
                left_branches=combined_parent_branches, right_branch=optimized_parent_branch
            )

            # If optimized_parent_branch couldn't be combined with any of the existing ones, add it to the list.
            if not any(x.combined_branch is not None for x in combination_results):
                combined_parent_branches.append(optimized_parent_branch)
            # Otherwise, replaced the branch with the one that was combined in combined_parent_branches
            else:
                combined_parent_branches = [
                    (
                        branch_combination_result.left_branch
                        if branch_combination_result.combined_branch is None
                        else branch_combination_result.combined_branch
                    )
                    for branch_combination_result in combination_results
                ]

        logger.debug(
            LazyFormat(
                "Possible branches combined.",
                count_of_branches_before_combination=len(optimized_parent_branches),
                count_of_branches_after_combination=len(combined_parent_branches),
            )
        )

        assert len(combined_parent_branches) > 0

        # If we were able to reduce the parent branches of the CombineAggregatedOutputsNode into a single one, there's
        # no need for a CombineAggregatedOutputsNode.
        if len(combined_parent_branches) == 1:
            result = OptimizeBranchResult(optimized_branch=combined_parent_branches[0])
        else:
            result = OptimizeBranchResult(
                optimized_branch=CombineAggregatedOutputsNode.create(parent_nodes=combined_parent_branches)
            )

        self._node_to_result[node] = result
        return result

    def visit_constrain_time_range_node(self, node: ConstrainTimeRangeNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_join_over_time_range_node(self, node: JoinOverTimeRangeNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_semi_additive_join_node(self, node: SemiAdditiveJoinNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_metric_time_dimension_transform_node(  # noqa: D102
        self, node: MetricTimeDimensionTransformNode
    ) -> OptimizeBranchResult:
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def optimize(self, dataflow_plan: DataflowPlan) -> DataflowPlan:  # noqa: D102
        optimized_result: OptimizeBranchResult = dataflow_plan.sink_node.accept(self)

        logger.debug(
            LazyFormat(
                "Optimized dataflow plan",
                original_plan=dataflow_plan.sink_node.structure_text(),
                optimized_plan=optimized_result.optimized_branch.structure_text(),
            )
        )

        return DataflowPlan(
            plan_id=DagId.from_id_prefix(StaticIdPrefix.OPTIMIZED_DATAFLOW_PLAN_PREFIX),
            sink_nodes=[optimized_result.optimized_branch],
        )

    def visit_join_to_time_spine_node(self, node: JoinToTimeSpineNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_add_generated_uuid_column_node(  # noqa: D102
        self, node: AddGeneratedUuidColumnNode
    ) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_join_conversion_events_node(self, node: JoinConversionEventsNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_join_to_custom_granularity_node(  # noqa: D102
        self, node: JoinToCustomGranularityNode
    ) -> OptimizeBranchResult:
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_min_max_node(self, node: MinMaxNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)

    def visit_alias_specs_node(self, node: AliasSpecsNode) -> OptimizeBranchResult:  # noqa: D102
        self._log_visit_node_type(node)
        return self._default_base_output_handler(node)



from __future__ import annotations

from functools import cached_property
from typing import Callable, Union

from typing_extensions import override

from metricflow_semantics.mf_logging.pretty_print import mf_pformat_dict


class LazyFormat:
    """Lazily formats the given objects into a string representation for logging.

    The formatting is done when this object is converted to a string to be compatible with the standard `logging`
    library. This is done lazily as the formatting operation can be expensive and allows debug log statements to not
    incur a performance overhead in production.

    Example:
        logger.debug(LazyFormat(lambda: LazyFormat("Found path", start_point=point_0, end_point=point_1)))

        ->

        DEBUG - Found path
            start_point: Point(x=0.0, y=0.0)
            end_point: Point(x=1.0, y=1.0)

    To aid migration of existing log statements, the message can be a function so that f-strings can be lazily
    evaluated as well. This style will be deprecated as log statements are updated.

    Example:
        logger.debug(LazyFormat(lambda: f"Result is: {expensive_function()}"))

        ->

        logger.debug(LazyFormat(lambda: f"Result is: {expensive_function()}")
    """

    def __init__(self, message: Union[str, Callable[[], str]], **kwargs) -> None:  # type: ignore[no-untyped-def]
        """Initializer.

        Args:
            message: The message or a function that returns a message.
            **kwargs: A dictionary of objects to format to text when `__str__` is called on this object.
        """
        self._message = message
        self._kwargs = kwargs

    @cached_property
    def _str_value(self) -> str:
        """Cache the result as `__str__` can be called multiple times for multiple log handlers."""
        if callable(self._message):
            message = self._message()
        else:
            message = self._message
        return mf_pformat_dict(message, self._kwargs, preserve_raw_strings=True)

    @override
    def __str__(self) -> str:
        return self._str_value



from __future__ import annotations

import collections
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from functools import cached_property
from typing import FrozenSet, Iterable, Optional, Sequence, Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.protocols.dimension import DimensionType
from dbt_semantic_interfaces.references import (
    DimensionReference,
    EntityReference,
    GroupByMetricReference,
    MetricReference,
    SemanticModelReference,
)
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from typing_extensions import override

from metricflow_semantics.assert_one_arg import assert_exactly_one_arg_set
from metricflow_semantics.model.linkable_element_property import LinkableElementProperty
from metricflow_semantics.model.semantic_model_derivation import SemanticModelDerivation
from metricflow_semantics.time.granularity import ExpandedTimeGranularity
from metricflow_semantics.workarounds.reference import sorted_semantic_model_references

logger = logging.getLogger(__name__)


class LinkableElementType(Enum):
    """Enumeration of the possible types of linkable element we are encountering or expecting.

    LinkableElements effectively map on to LinkableSpecs and queryable semantic manifest elements such
    as Metrics, Dimensions, and Entities. This provides the full set of types we might encounter, and is
    useful for ensuring that we are always getting the correct LinkableElement from a given part of the
    codebase - e.g., to ensure we are not accidentally getting an Entity when we expect a Dimension.
    """

    DIMENSION = "dimension"
    ENTITY = "entity"
    METRIC = "metric"
    TIME_DIMENSION = "time_dimension"

    @property
    def is_dimension_type(self) -> bool:
        """Property to simplify scenarios where callers need to know whether or not this represents a dimension."""
        # Use a local alias to allow type refinement for the static exhaustive switch assertion
        element_type = self
        if element_type is LinkableElementType.DIMENSION or element_type is LinkableElementType.TIME_DIMENSION:
            return True
        elif element_type is LinkableElementType.ENTITY or element_type is LinkableElementType.METRIC:
            return False
        else:
            return assert_values_exhausted(element_type)


@dataclass(frozen=True)
class ElementPathKey:
    """A key that can uniquely identify an element and the joins used to realize the element."""

    element_name: str
    element_type: LinkableElementType
    entity_links: Tuple[EntityReference, ...]
    time_granularity: Optional[ExpandedTimeGranularity] = None
    date_part: Optional[DatePart] = None
    metric_subquery_entity_links: Tuple[EntityReference, ...] = ()

    def __post_init__(self) -> None:
        """Asserts all requirements associated with the element_type are met."""
        element_type = self.element_type
        if element_type is LinkableElementType.TIME_DIMENSION:
            assert (
                self.time_granularity
            ), "Time granularity must be specified for all ElementPathKeys associated with time dimensions!"
        elif (
            element_type is LinkableElementType.DIMENSION
            or element_type is LinkableElementType.ENTITY
            or element_type is LinkableElementType.METRIC
        ):
            pass
        else:
            assert_values_exhausted(element_type)

        assert len(set(self.entity_links)) == len(
            self.entity_links
        ), f"Duplicate found in `entity_links`: {self.entity_links}."


@dataclass(frozen=True)
class SemanticModelJoinPathElement(SerializableDataclass):
    """Describes joining a semantic model by the given entity."""

    semantic_model_reference: SemanticModelReference
    join_on_entity: EntityReference


@dataclass(frozen=True)
class LinkableElement(SemanticModelDerivation, SerializableDataclass, ABC):
    """An entity / dimension that may have been joined by entities."""

    properties: Tuple[LinkableElementProperty, ...]

    def __post_init__(self) -> None:  # noqa: D105
        if len(self.property_set) != len(self.properties):
            duplicate_properties = [item for item, count in collections.Counter(self.properties).items() if count > 1]
            assert False, f"Found duplicate properties {duplicate_properties} in {self.properties}"

        assert self.properties == tuple(sorted(self.properties)), f"Properties are not sorted: {self.properties}"

    @cached_property
    def property_set(self) -> FrozenSet[LinkableElementProperty]:  # noqa: D102
        return frozenset(self.properties)

    @property
    @abstractmethod
    def element_type(self) -> LinkableElementType:
        """The LinkableElementType describing what this instance represents."""
        raise NotImplementedError

    @property
    @abstractmethod
    def semantic_model_origin(self) -> SemanticModelReference:
        """The semantic model where this element was defined, if one exists.

        If no such model exists, the element will return the VIRTUAL_SEMANTIC_MODEL_REFERENCE, as it is
        either a virtual construct (e.g., metric_time) or a composite of semantic model inputs that could be used as
        a virtual semantic model at some point (e.g., metric queries requested as filter inputs).
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def as_union(self) -> LinkableElementUnion:
        """Return `self` in a union-type container for better serialization support."""
        raise NotImplementedError


@dataclass(frozen=True)
class LinkableDimension(LinkableElement, SerializableDataclass):
    """Describes how a dimension can be realized by joining based on entity links."""

    # The semantic model where this dimension was defined.
    defined_in_semantic_model: Optional[SemanticModelReference]
    element_name: str
    dimension_type: DimensionType
    entity_links: Tuple[EntityReference, ...]
    join_path: SemanticModelJoinPath
    time_granularity: Optional[ExpandedTimeGranularity]
    date_part: Optional[DatePart]

    @staticmethod
    def create(  # noqa: D102
        properties: Iterable[LinkableElementProperty],
        defined_in_semantic_model: Optional[SemanticModelReference],
        element_name: str,
        dimension_type: DimensionType,
        entity_links: Tuple[EntityReference, ...],
        join_path: SemanticModelJoinPath,
        time_granularity: Optional[ExpandedTimeGranularity],
        date_part: Optional[DatePart],
    ) -> LinkableDimension:
        return LinkableDimension(
            properties=tuple(sorted(set(properties))),
            defined_in_semantic_model=defined_in_semantic_model,
            element_name=element_name,
            dimension_type=dimension_type,
            entity_links=entity_links,
            join_path=join_path,
            time_granularity=time_granularity,
            date_part=date_part,
        )

    @property
    @override
    def element_type(self) -> LinkableElementType:
        return (
            LinkableElementType.DIMENSION
            if self.dimension_type is DimensionType.CATEGORICAL
            else LinkableElementType.TIME_DIMENSION
        )

    @property
    def path_key(self) -> ElementPathKey:  # noqa: D102
        return ElementPathKey(
            element_name=self.element_name,
            element_type=self.element_type,
            entity_links=self.entity_links,
            time_granularity=self.time_granularity,
            date_part=self.date_part,
        )

    @property
    def reference(self) -> DimensionReference:  # noqa: D102
        return DimensionReference(element_name=self.element_name)

    @property
    @override
    def derived_from_semantic_models(self) -> Sequence[SemanticModelReference]:
        semantic_model_references = set()
        if self.defined_in_semantic_model:
            semantic_model_references.add(self.defined_in_semantic_model)
        semantic_model_references.update(self.join_path.derived_from_semantic_models)

        return sorted_semantic_model_references(semantic_model_references)

    @property
    @override
    def semantic_model_origin(self) -> SemanticModelReference:
        """Returns the semantic model reference pointing to the model where the dimension is defined.

        For virtual dimensions, such as metric_time, where there is no semantic model definition we return the
        virtual semantic model reference.
        """
        return (
            self.defined_in_semantic_model
            if self.defined_in_semantic_model
            else SemanticModelDerivation.VIRTUAL_SEMANTIC_MODEL_REFERENCE
        )

    @property
    @override
    def as_union(self) -> LinkableElementUnion:
        return LinkableElementUnion(linkable_dimension=self)


@dataclass(frozen=True)
class LinkableEntity(LinkableElement, SerializableDataclass):
    """Describes how an entity can be realized by joining based on entity links."""

    # The semantic model where this entity was defined.
    defined_in_semantic_model: SemanticModelReference
    element_name: str
    entity_links: Tuple[EntityReference, ...]
    join_path: SemanticModelJoinPath

    @staticmethod
    def create(  # noqa: D102
        properties: Iterable[LinkableElementProperty],
        defined_in_semantic_model: SemanticModelReference,
        element_name: str,
        entity_links: Tuple[EntityReference, ...],
        join_path: SemanticModelJoinPath,
    ) -> LinkableEntity:
        return LinkableEntity(
            properties=tuple(sorted(set(properties))),
            defined_in_semantic_model=defined_in_semantic_model,
            element_name=element_name,
            entity_links=entity_links,
            join_path=join_path,
        )

    @property
    @override
    def element_type(self) -> LinkableElementType:
        return LinkableElementType.ENTITY

    @property
    def path_key(self) -> ElementPathKey:  # noqa: D102
        return ElementPathKey(
            element_name=self.element_name, element_type=self.element_type, entity_links=self.entity_links
        )

    @property
    def reference(self) -> EntityReference:  # noqa: D102
        return EntityReference(element_name=self.element_name)

    @property
    @override
    def derived_from_semantic_models(self) -> Sequence[SemanticModelReference]:
        semantic_model_references = {self.defined_in_semantic_model}
        semantic_model_references.update(self.join_path.derived_from_semantic_models)
        return sorted_semantic_model_references(semantic_model_references)

    @property
    @override
    def semantic_model_origin(self) -> SemanticModelReference:
        return self.defined_in_semantic_model

    @property
    @override
    def as_union(self) -> LinkableElementUnion:
        return LinkableElementUnion(linkable_entity=self)


@dataclass(frozen=True)
class LinkableMetric(LinkableElement, SerializableDataclass):
    """Describes how a metric can be realized by joining based on entity links."""

    join_path: SemanticModelToMetricSubqueryJoinPath

    @staticmethod
    def create(  # noqa: D102
        properties: Iterable[LinkableElementProperty], join_path: SemanticModelToMetricSubqueryJoinPath
    ) -> LinkableMetric:
        return LinkableMetric(
            properties=tuple(sorted(set(properties))),
            join_path=join_path,
        )

    def __post_init__(self) -> None:
        """Ensure expected LinkableElementProperties have been set.

        LinkableMetrics always require a join to a metric subquery.
        """
        super().__post_init__()
        assert {LinkableElementProperty.METRIC, LinkableElementProperty.JOINED}.issubset(self.property_set)

    @property
    @override
    def element_type(self) -> LinkableElementType:
        return LinkableElementType.METRIC

    @property
    def element_name(self) -> str:  # noqa: D102
        return self.reference.element_name

    @property
    def path_key(self) -> ElementPathKey:  # noqa: D102
        return ElementPathKey(
            element_name=self.element_name,
            element_type=self.element_type,
            entity_links=self.join_path.entity_links,
            metric_subquery_entity_links=self.metric_subquery_entity_links,
        )

    @property
    def metric_reference(self) -> MetricReference:  # noqa: D102
        return self.join_path.metric_subquery_join_path_element.metric_reference

    @property
    def reference(self) -> GroupByMetricReference:  # noqa: D102
        return GroupByMetricReference(self.metric_reference.element_name)

    @property
    @override
    def derived_from_semantic_models(self) -> Sequence[SemanticModelReference]:
        """Semantic models needed to build and join to this LinkableMetric.

        Includes semantic models used in the join paths for both the inner and outer queries (if applicable),
        plus the semantic models the metric's measure(s) originated from.
        """
        semantic_model_references = set(self.join_path.metric_subquery_join_path_element.derived_from_semantic_models)
        if self.join_path.semantic_model_join_path:
            semantic_model_references.update(self.join_path.semantic_model_join_path.derived_from_semantic_models)
        if self.metric_to_entity_join_path:
            semantic_model_references.update(self.metric_to_entity_join_path.derived_from_semantic_models)

        return sorted_semantic_model_references(semantic_model_references)

    @property
    @override
    def semantic_model_origin(self) -> SemanticModelReference:
        """Returns the virtual semantic model reference, as metrics are not defined in semantic models.

        Metrics may be used as virtual source nodes for metrics as dimensions use cases, and we represent that here.
        """
        return SemanticModelDerivation.VIRTUAL_SEMANTIC_MODEL_REFERENCE

    @property
    def metric_to_entity_join_path(self) -> Optional[SemanticModelJoinPath]:
        """Join path used in metric subquery to join entity to metric, if needed."""
        return self.join_path.metric_subquery_join_path_element.metric_to_entity_join_path

    @property
    def metric_subquery_entity_links(self) -> Tuple[EntityReference, ...]:
        """Entity links used to join the metric to the entity it's grouped by in the metric subquery.

        Includes the `join_on_entity`, which will always be the last entity link.
        """
        return self.join_path.metric_subquery_entity_links

    @property
    @override
    def as_union(self) -> LinkableElementUnion:
        return LinkableElementUnion(linkable_metric=self)


@dataclass(frozen=True)
class LinkableElementUnion(SerializableDataclass):
    """A union type to use in classes that require a concrete implementation for serialization."""

    linkable_dimension: Optional[LinkableDimension] = None
    linkable_entity: Optional[LinkableEntity] = None
    linkable_metric: Optional[LinkableMetric] = None

    def __post_init__(self) -> None:  # noqa: D105
        assert_exactly_one_arg_set(
            linkable_dimension=self.linkable_dimension,
            linkable_entity=self.linkable_entity,
            linkable_metric=self.linkable_metric,
        )

    @property
    def linkable_element(self) -> LinkableElement:  # noqa: D102
        if self.linkable_dimension is not None:
            return self.linkable_dimension
        elif self.linkable_entity is not None:
            return self.linkable_entity
        elif self.linkable_metric is not None:
            return self.linkable_metric

        assert False, "All fields are None - this should have been caught in object initialization."


@dataclass(frozen=True)
class SemanticModelJoinPath(SemanticModelDerivation, SerializableDataclass):
    """Describes a series of joins between the measure semantic model, and other semantic models by entity.

    For example:

    (measure_source JOIN dimension_source0 ON entity A) JOIN dimension_source1 ON entity B

    would be represented by 2 path elements [(semantic_model0, A), (dimension_source1, B)]
    """

    left_semantic_model_reference: SemanticModelReference
    path_elements: Tuple[SemanticModelJoinPathElement, ...] = ()

    @property
    def last_path_element(self) -> SemanticModelJoinPathElement:  # noqa: D102
        assert len(self.path_elements) > 0
        return self.path_elements[-1]

    @property
    def last_semantic_model_reference(self) -> SemanticModelReference:
        """The last semantic model that would be joined in this path."""
        return self.last_path_element.semantic_model_reference

    @property
    def last_entity_link(self) -> EntityReference:  # noqa: D102
        return self.last_path_element.join_on_entity

    @property
    def entity_links(self) -> Tuple[EntityReference, ...]:  # noqa: D102
        return tuple(path_element.join_on_entity for path_element in self.path_elements)

    @staticmethod
    def from_single_element(
        left_semantic_model_reference: SemanticModelReference,
        right_semantic_model_reference: SemanticModelReference,
        join_on_entity: EntityReference,
    ) -> SemanticModelJoinPath:
        """Build SemanticModelJoinPath with just one join path element."""
        return SemanticModelJoinPath(
            left_semantic_model_reference=left_semantic_model_reference,
            path_elements=(
                SemanticModelJoinPathElement(
                    semantic_model_reference=right_semantic_model_reference,
                    join_on_entity=join_on_entity,
                ),
            ),
        )

    @property
    def derived_from_semantic_models(self) -> Sequence[SemanticModelReference]:
        """Unique semantic models used in this join path."""
        semantic_model_references = set()
        semantic_model_references.add(self.left_semantic_model_reference)
        for path_element in self.path_elements:
            semantic_model_references.add(path_element.semantic_model_reference)

        return sorted_semantic_model_references(semantic_model_references)


@dataclass(frozen=True)
class MetricSubqueryJoinPathElement(SerializableDataclass):
    """Describes joining from a semantic model to a metric subquery.

    Args:
        metric_reference: The metric that's aggregated in the subquery.
        derived_from_semantic_models: The semantic models that the measure's input metrics are defined in.
        join_on_entity: The entity that the metric is grouped by in the subquery. This will be updated in V2 to allow a list
            of entities & dimensions.
        entity_links: Sequence of entities joined to get from a metric source to the `join_on_entity`. Should not include
            the `join_on_entity`.
        metric_to_entity_join_path: Describes the join path used in the subquery to join the metric to the `join_on_entity`.
            Can be none if all required elements are defined in the same semantic model.
    """

    metric_reference: MetricReference
    derived_from_semantic_models: Tuple[SemanticModelReference, ...]
    join_on_entity: EntityReference
    entity_links: Tuple[EntityReference, ...]
    metric_to_entity_join_path: Optional[SemanticModelJoinPath] = None

    def __post_init__(self) -> None:  # noqa: D105
        assert (
            self.derived_from_semantic_models
        ), "There must be at least one semantic model from which the metric is derived."


@dataclass(frozen=True)
class SemanticModelToMetricSubqueryJoinPath(SerializableDataclass):
    """Describes how to join from a semantic model to a metric subquery.

    Starts with semantic model join path, if needed. Always ends with metric subquery join path.
    """

    metric_subquery_join_path_element: MetricSubqueryJoinPathElement
    semantic_model_join_path: SemanticModelJoinPath

    @property
    def entity_links(self) -> Tuple[EntityReference, ...]:  # noqa: D102
        return (self.semantic_model_join_path.entity_links if self.semantic_model_join_path else ()) + (
            self.metric_subquery_join_path_element.join_on_entity,
        )

    @property
    def metric_subquery_entity_links(self) -> Tuple[EntityReference, ...]:  # noqa: D102
        return self.metric_subquery_join_path_element.entity_links + (
            self.metric_subquery_join_path_element.join_on_entity,
        )



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Sequence

from metricflow.telemetry.models import FunctionEndEvent, FunctionStartEvent, TelemetryPayload

PayloadType = Dict[Any, Any]  # type: ignore
logger = logging.getLogger(__name__)


class TelemetryHandler(ABC):
    """Base class to record telemetry to some destination."""

    @abstractmethod
    def _write_log(self, client_id: str, payload: PayloadType) -> None:
        """Subclasses should implement this to log a serialized version of the payload."""
        pass

    def log(
        self,
        client_id: str,
        function_start_event: Optional[FunctionStartEvent] = None,
        function_end_event: Optional[FunctionEndEvent] = None,
    ) -> bool:
        """Log an event to telemetry."""
        payload = TelemetryPayload(
            client_id=client_id,
            function_start_events=(function_start_event,) if function_start_event else (),
            function_end_events=(function_end_event,) if function_end_event else (),
        )
        self._write_log(client_id, payload.dict())
        return True


class ToMemoryTelemetryHandler(TelemetryHandler):
    """Records telemetry events to memory for testing purposes."""

    def __init__(self) -> None:  # noqa: D107
        self._payloads: List[TelemetryPayload] = []

    def _write_log(self, client_id: str, payload: PayloadType) -> None:
        pass

    @property
    def payloads(self) -> Sequence[TelemetryPayload]:  # noqa: D102
        return self._payloads

    def log(
        self,
        client_id: str,
        function_start_event: Optional[FunctionStartEvent] = None,
        function_end_event: Optional[FunctionEndEvent] = None,
    ) -> bool:
        """Log an event to telemetry."""
        payload = TelemetryPayload(
            client_id=client_id,
            function_start_events=(function_start_event,) if function_start_event else (),
            function_end_events=(function_end_event,) if function_end_event else (),
        )
        if len(self._payloads) > 10:
            self._payloads.pop()

        self._payloads.append(payload)
        return True



from __future__ import annotations

import pathlib
from typing import Dict, Optional

from dbt_semantic_interfaces.implementations.semantic_manifest import PydanticSemanticManifest
from dbt_semantic_interfaces.parsing.dir_to_model import (
    parse_directory_of_yaml_files_to_semantic_manifest,
)
from dbt_semantic_interfaces.validations.semantic_manifest_validator import SemanticManifestValidator


def load_semantic_manifest(
    yaml_file_directory: pathlib.Path,
    template_mapping: Optional[Dict[str, str]] = None,
) -> PydanticSemanticManifest:
    """Reads the manifest YAMLs from the standard location, applies transformations, runs validations."""
    try:
        build_result = parse_directory_of_yaml_files_to_semantic_manifest(
            str(yaml_file_directory), template_mapping=template_mapping
        )
        validator = SemanticManifestValidator[PydanticSemanticManifest]()
        validator.checked_validations(build_result.semantic_manifest)
        return build_result.semantic_manifest
    except Exception as e:
        raise RuntimeError(f"Error while loading semantic manifest: {yaml_file_directory}") from e



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence, Tuple

from dbt_semantic_interfaces.references import MetricReference
from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput


@dataclass(frozen=True)
class DuplicateMetricIssue(MetricFlowQueryResolutionIssue):
    """Describes when there are duplicate metrics in a query."""

    duplicate_metric_references: Tuple[MetricReference, ...]

    @staticmethod
    def from_parameters(  # noqa: D102
        duplicate_metric_references: Sequence[MetricReference],
        query_resolution_path: MetricFlowQueryResolutionPath,
    ) -> DuplicateMetricIssue:
        return DuplicateMetricIssue(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=(),
            duplicate_metric_references=tuple(duplicate_metric_references),
            query_resolution_path=query_resolution_path,
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        return (
            f"Query contains duplicate metrics: "
            f"{[metric_reference.element_name for metric_reference in self.duplicate_metric_references]}"
        )

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> DuplicateMetricIssue:
        return DuplicateMetricIssue(
            issue_type=self.issue_type,
            parent_issues=tuple(issue.with_path_prefix(path_prefix) for issue in self.parent_issues),
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
            duplicate_metric_references=self.duplicate_metric_references,
        )



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass

from typing_extensions import override

from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput
from metricflow_semantics.specs.instance_spec import InstanceSpec


@dataclass(frozen=True)
class ScdRequiresMetricTimeIssue(MetricFlowQueryResolutionIssue):
    """Describes an issue with a query that includes a SCD group by but does not include metric_time."""

    scds_in_query: Sequence[InstanceSpec]

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        dim_str = ", ".join(f"'{scd.qualified_name}'" for scd in self.scds_in_query)
        return (
            "Your query contains the following group bys, which are SCDs or contain SCDs "
            f"in the join path: [{dim_str}].\n\nA query containing SCDs must also contain "
            "the metric_time dimension in order to join the SCD table to the valid time "
            "range. Please add metric_time to the query and try again. If you're "
            "using agg_time_dimension, use metric_time instead."
        )

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> ScdRequiresMetricTimeIssue:
        return ScdRequiresMetricTimeIssue(
            issue_type=self.issue_type,
            parent_issues=self.parent_issues,
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
            scds_in_query=self.scds_in_query,
        )

    @staticmethod
    def from_parameters(  # noqa: D102
        scds_in_query: Sequence[InstanceSpec], query_resolution_path: MetricFlowQueryResolutionPath
    ) -> ScdRequiresMetricTimeIssue:
        return ScdRequiresMetricTimeIssue(
            issue_type=MetricFlowQueryIssueType.ERROR,
            parent_issues=(),
            query_resolution_path=query_resolution_path,
            scds_in_query=scds_in_query,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.specs.where_filter.where_filter_spec import WhereFilterSpec
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class WhereConstraintNode(DataflowPlanNode):
    """Remove rows using a WHERE clause.

    Attributes:
        where_specs: Specifications for the WHERE clause to filter rows.
        always_apply: Indicator if the WHERE clause should always be applied.
    """

    where_specs: Sequence[WhereFilterSpec]
    always_apply: bool

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(  # noqa: D102
        parent_node: DataflowPlanNode,
        where_specs: Sequence[WhereFilterSpec],
        always_apply: bool = False,
    ) -> WhereConstraintNode:
        return WhereConstraintNode(
            parent_nodes=(parent_node,),
            where_specs=where_specs,
            always_apply=always_apply,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_WHERE_CONSTRAINT_ID_PREFIX

    @property
    def where(self) -> WhereFilterSpec:
        """Returns the specs for the elements that it should pass."""
        return WhereFilterSpec.merge_iterable(self.where_specs)

    @property
    def input_where_specs(self) -> Sequence[WhereFilterSpec]:
        """Returns the discrete set of input where filter specs for this node.

        This is useful for things like predicate pushdown, where we need to differentiate between individual specs
        for pushdown operations on the filter spec level. We merge them for things like rendering and node comparisons,
        but in some cases we may be able to push down a subset of the input specs for efficiency reasons.
        """
        return self.where_specs

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_where_constraint_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        # Can't put the where condition here as it can cause rendering issues when there are SQL execution parameters.
        # e.g. "Constrain Output with WHERE listing__country = :1"
        return "Constrain Output with WHERE"

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        properties = tuple(super().displayed_properties) + (DisplayedProperty("where_condition", self.where),)
        if self.always_apply:
            properties = properties + (DisplayedProperty("All filters always applied:", self.always_apply),)
        return properties

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return (
            isinstance(other_node, self.__class__)
            and other_node.where == self.where
            and other_node.always_apply == self.always_apply
        )

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> WhereConstraintNode:  # noqa: D102
        assert len(new_parent_nodes) == 1
        return WhereConstraintNode.create(
            parent_node=new_parent_nodes[0],
            where_specs=self.input_where_specs,
            always_apply=self.always_apply,
        )
